# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/API_DataSource.ipynb.

# %% auto 0
__all__ = ['DataSource']

# %% ../../notebooks/API_DataSource.ipynb 4
from typing import *

# %% ../../notebooks/API_DataSource.ipynb 5
import os

import pandas as pd
from fastcore.foundation import patch
from datetime import datetime, timedelta
from pathlib import Path
import requests

from airt.logger import get_logger, set_level
from airt.helper import (
    get_data,
    post_data,
    delete_data,
    add_ready_column,
    get_values_from_item,
    generate_df,
    dict_to_df,
    add_example_to_docs,
)


from airt.components.client import Client
from airt.components.progress_status import ProgressStatus
from airt.components.model import Model

# %% ../../notebooks/API_DataSource.ipynb 7
logger = get_logger(__name__)

# %% ../../notebooks/API_DataSource.ipynb 10
class DataSource:
    """A class for managing datasources and training ML models on them.

    To instantiate the **DataSource** class, please call `DataBlob.to_datasource` method of the `DataBlob` class.

    The **DataSource** class has two categories of methods,

    * Methods for managing the datasources.
    * Method for training a model against a datasource.

    Methods such as `delete`, `ls`, `details`, `head`, etc., can be used to manage and obtain additional information from a datasource instance.

    And, the `train` method can be used to train a new model against a datasource instance.

    All the function calls to the library are asynchronous and they return immediately. To manage completion, methods inside the returned object
    will return a status object indicating the completion status and a method to display an interactive progress bar that can be called to check the progress.
    """

    BASIC_DS_COLS = [
        "uuid",
        "datablob",
        "region",
        "cloud_provider",
        "tags",
        "pulled_on",
        "completed_steps",
        "total_steps",
        "no_of_rows",
        "folder_size",
    ]

    ALL_DS_COLS = BASIC_DS_COLS + ["user", "error", "disabled"]

    COLS_TO_RENAME = {
        "uuid": "datasource_uuid",
        "datablob": "datablob_uuid",
        "user": "user_uuid",
    }

    def __init__(
        self,
        uuid: str,
        datablob: Optional[str] = None,
        folder_size: Optional[int] = None,
        no_of_rows: Optional[int] = None,
        error: Optional[str] = None,
        disabled: Optional[bool] = None,
        created: Optional[str] = None,
        pulled_on: Optional[str] = None,
        user: Optional[str] = None,
        hash: Optional[str] = None,
        region: Optional[str] = None,
        cloud_provider: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        total_steps: Optional[int] = None,
        completed_steps: Optional[int] = None,
    ):
        """Constructs a new `DataSource` instance.

        Warning:
            Do not construct this object directly by calling the constructor, please use `DataBlob.to_datasource` method instead.

        Args:
            uuid: DataSource uuid.
            datablob: Datablob uuid.
            folder_size: The uploaded datasource's size in bytes.
            no_of_rows: The number of records in the datasource.
            error: Contains the error message if the processing of the datasource fails.
            disabled: A flag that indicates the datasource's status. If the datasource is deleted, then **False** will be set.
            created: The datasource creation date.
            pulled_on: The most recent date the datasource was uploaded.
            user: The uuid of the user who created the datasource.
            hash: The datasource hash.
            region: The region name of the cloud provider where the datasource is stored
            cloud_provider: The name of the cloud storage provider where the datasource is stored.
            tags: Tag names associated with the datasource.
            total_steps: The number of steps required to upload the datasource to the server.
            completed_steps: The number of steps completed during the datasource's upload to the server.
        """
        self.uuid = uuid
        self.datablob = datablob
        self.folder_size = folder_size
        self.no_of_rows = no_of_rows
        self.error = error
        self.disabled = disabled
        self.created = created
        self.pulled_on = pulled_on
        self.user = user
        self.hash = hash
        self.region = region
        self.cloud_provider = cloud_provider
        self.tags = tags
        self.total_steps = total_steps
        self.completed_steps = completed_steps

    @property
    def dtypes(self) -> pd.DataFrame:
        """Return the dtypes of the datasource.

        Returns:
            A pandas DataFrame containing the data type of each column.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
        """
        dtypes = Client._get_data(relative_url=f"/datasource/{self.uuid}/dtypes")
        return pd.DataFrame([dtypes])

    @staticmethod
    def ls(
        offset: int = 0,
        limit: int = 100,
        disabled: bool = False,
        completed: bool = False,
    ) -> List["DataSource"]:
        """Return the list of `DataSource` instances available in server.

        Args:
            offset: The number of datasources to offset at the beginning. If **None**,
                then the default value **0** will be used.
            limit: The maximum number of datasources to return from the server. If **None**,
                then the default value **100** will be used.
            disabled: If set to **True**, then only the deleted datasources will be returned.
                Else, the default value **False** will be used to return only the list
                of active datasources.
            completed: If set to **True**, then only the datasources that are successfully processed
                in server will be returned. Else, the default value **False** will be used to
                return all the datasources.

        Returns:
            A list of `DataSource` instances available in server.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
        """
        lists = Client._get_data(
            relative_url=f"/datasource/?disabled={disabled}&completed={completed}&offset={offset}&limit={limit}"
        )

        dsx = [
            DataSource(
                uuid=ds["uuid"],
                datablob=ds["datablob"],
                folder_size=ds["folder_size"],
                no_of_rows=ds["no_of_rows"],
                region=ds["region"],
                cloud_provider=ds["cloud_provider"],
                error=ds["error"],
                disabled=ds["disabled"],
                created=ds["created"],
                pulled_on=ds["pulled_on"],
                user=ds["user"],
                hash=ds["hash"],
                tags=ds["tags"],
                total_steps=ds["total_steps"],
                completed_steps=ds["completed_steps"],
            )
            for ds in lists
        ]

        return dsx

    @staticmethod
    def as_df(dsx: List["DataSource"]) -> pd.DataFrame:
        """Return the details of `DataSource` instances as a pandas dataframe.

        Args:
            dsx: List of `DataSource` instances.

        Returns:
            Details of the datasources in a dataframe.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
        """

        ds_lists = [{i: getattr(ds, i) for i in DataSource.ALL_DS_COLS} for ds in dsx]

        for ds in ds_lists:
            ds["tags"] = get_values_from_item(ds["tags"], "name")

        lists_df = generate_df(ds_lists, DataSource.BASIC_DS_COLS)
        df = add_ready_column(lists_df)

        df = df.rename(columns=DataSource.COLS_TO_RENAME)

        return df

    def is_ready(self):
        raise NotImplementedError()

    def progress_bar(self, sleep_for: Union[int, float] = 5, timeout: int = 0):
        raise NotImplementedError()

    def wait(self, sleep_for: Union[int, float] = 1, timeout: int = 0):
        raise NotImplementedError()

    def delete(self) -> pd.DataFrame:
        raise NotImplementedError()

    def details(self) -> pd.DataFrame:
        raise NotImplementedError()

    def tag(self, name: str) -> pd.DataFrame:
        raise NotImplementedError()

    def head(self) -> pd.DataFrame:
        raise NotImplementedError()

    def train(
        self,
        *,
        client_column: str,
        timestamp_column: Optional[str] = None,
        target_column: str,
        target: str,
        predict_after: timedelta,
    ) -> "airt.client.Model":  # type: ignore
        raise NotImplementedError()

# %% ../../notebooks/API_DataSource.ipynb 11
def _docstring_example():
    """
    Example:
        ```python
        # Importing necessary libraries
        from datetime import timedelta

        from  airt.client import Client, DataBlob, DataSource

        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a datablob
        # In this example, the datablob will be stored in an AWS S3 bucket. The
        # access_key and the secret_key are set in the AWS_ACCESS_KEY_ID and
        # AWS_SECRET_ACCESS_KEY environment variables, and the region is set to
        # eu-west-3; feel free to change the cloud provider and the region to
        # suit your needs.
        db = DataBlob.from_s3(
            uri="{fill in uri}",
            cloud_provider="aws",
            region="eu-west-3"
        )

        # Display the status in a progress bar
        db.progress_bar()

        # Create a datasource
        ds = db.to_datasource(
            file_type="{fill in file_type}",
            index_column="{fill in index_column}",
            sort_by="{fill in sort_by}",
        )

        # Display the status in a progress bar
        # Call the wait method to wait for the progress to finish but
        # without displaying an interactive progress bar.
        ds.progress_bar()

        # Display the ready status
        print(ds.is_ready())

        # Display the data types of the datasource's columns.
        print(ds.dtypes)

        # Display the details of the datasource
        print(ds.details())

        # Display the details of all datasource created by the currently
        # logged-in user
        print(DataSource.as_df(DataSource.ls()))

        # Display the first few records of the datasource
        print(ds.head())

        # Train a model against the datasource.
        # This example predicts which users will perform a purchase
        # event ("*purchase") three hours before they actually do it.
        model = ds.train(
            client_column="{fill in client_column}",
            target_column="{fill in target_column}",
            target="*purchase",
            predict_after=timedelta(hours=3)
        )

        # Display the training status in a progress bar
        model.progress_bar()

        # Display the details of the newly created model
        print(model.details())

        # Tag the datasource
        print(ds.tag(name="{fill in tag_name}"))

        # Delete the datasource
        print(ds.delete())
        ```
    """
    pass

# %% ../../notebooks/API_DataSource.ipynb 13
add_example_to_docs(DataSource, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataSource.dtypes, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataSource.ls, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataSource.as_df, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 15
@patch
def is_ready(
    self: DataSource,
) -> bool:
    """Check if the method's progress is complete.

    Returns:
        **True** if the progress is completed, else **False**.
    """
    progress_status = ProgressStatus(relative_url=f"/datasource/{self.uuid}")

    return progress_status.is_ready()

# %% ../../notebooks/API_DataSource.ipynb 16
add_example_to_docs(DataSource.is_ready, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 18
@patch
def progress_bar(self: DataSource, sleep_for: Union[int, float] = 5, timeout: int = 0):
    """Blocks the execution and displays a progress bar showing the remote action progress.

    Args:
        sleep_for: The time interval in seconds between successive API calls.
        timeout: The maximum time allowed in seconds for the asynchronous call to complete. If not the
            progressbar will be terminated.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
        TimeoutError: in case of connection timeout.
    """
    progress_status = ProgressStatus(
        relative_url=f"/datasource/{self.uuid}", sleep_for=sleep_for, timeout=timeout
    )

    progress_status.progress_bar()

# %% ../../notebooks/API_DataSource.ipynb 19
add_example_to_docs(DataSource.progress_bar, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 20
@patch
def wait(self: DataSource, sleep_for: Union[int, float] = 1, timeout: int = 0):
    """Blocks execution while waiting for the remote action to complete.

    Args:
        sleep_for: The time interval in seconds between successive API calls.
        timeout: The maximum time allowed in seconds for the asynchronous call to complete. If not the
            progressbar will be terminated.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
        TimeoutError: in case of timeout.
    """

    progress_status = ProgressStatus(
        relative_url=f"/datasource/{self.uuid}", sleep_for=sleep_for, timeout=timeout
    )

    progress_status.wait()

# %% ../../notebooks/API_DataSource.ipynb 21
add_example_to_docs(DataSource.wait, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 27
@patch
def delete(self: DataSource) -> pd.DataFrame:
    """Delete a datasource from the server.

    Returns:
        A pandas DataFrame encapsulating the details of the deleted datasource.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """

    response = Client._delete_data(relative_url=f"/datasource/{self.uuid}")

    response["tags"] = get_values_from_item(response["tags"], "name")

    df = pd.DataFrame([response])[DataSource.BASIC_DS_COLS]

    df = df.rename(columns=DataSource.COLS_TO_RENAME)

    return add_ready_column(df)

# %% ../../notebooks/API_DataSource.ipynb 28
add_example_to_docs(DataSource.delete, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 31
@patch
def details(self: DataSource) -> pd.DataFrame:
    """Return details of a datasource.

    Returns:
        The datasource details as a pandas dataframe.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """

    response = Client._get_data(relative_url=f"/datasource/{self.uuid}")

    response["tags"] = get_values_from_item(response["tags"], "name")

    df = pd.DataFrame([response])[DataSource.ALL_DS_COLS]

    df = df.rename(columns=DataSource.COLS_TO_RENAME)

    return add_ready_column(df)

# %% ../../notebooks/API_DataSource.ipynb 32
add_example_to_docs(DataSource.details, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 34
@patch
def tag(self: DataSource, name: str) -> pd.DataFrame:
    """Tag an existing datasource in server.

    Args:
        name: A string to tag the datasource.

    Returns:
        A pandas dataframe with the details of the tagged datasource.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """
    response = Client._post_data(
        relative_url=f"/datasource/{self.uuid}/tag", json=dict(name=name)
    )

    response["tags"] = get_values_from_item(response["tags"], "name")

    df = pd.DataFrame([response])[DataSource.BASIC_DS_COLS]
    df = df.rename(columns=DataSource.COLS_TO_RENAME)

    return add_ready_column(df)

# %% ../../notebooks/API_DataSource.ipynb 35
add_example_to_docs(DataSource.tag, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 37
@patch
def head(self: DataSource) -> pd.DataFrame:
    """Return the first few rows of the datasource.

    Returns:
        The first few rows of the datasource as a pandas dataframe.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """
    response = Client._get_data(relative_url=f"/datasource/{self.uuid}/head")
    df = dict_to_df(response)

    return df

# %% ../../notebooks/API_DataSource.ipynb 38
add_example_to_docs(DataSource.head, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataSource.ipynb 40
@patch
def train(
    self: DataSource,
    *,
    client_column: str,
    timestamp_column: Optional[str] = None,
    target_column: str,
    target: str,
    predict_after: timedelta,
) -> Model:
    """Train a model against the datasource.

    This method trains the model for predicting which clients are most likely to have a specified
    event in the future.

    The call to this method is asynchronous and the progress can be checked using the progress bar method
    or the status flag attribute available in the `DataSource` class.

    For more model specific information, please check the documentation of `Model` class.

    Args:
        client_column: The column name that uniquely identifies the users/clients.
        timestamp_column: The timestamp column indicating the time of an event. If not passed,
            then the default value **None** will be used.
        target_column: Target column name that indicates the type of the event.
        target: Target event name to train and make predictions. You can pass the target event as a string or as a
            regular expression for predicting more than one event. For example, passing ***checkout** will
            train a model to predict any checkout event.
        predict_after: Time delta in hours of the expected target event.

    Returns:
        An instance of the `Model` class.

    Raises:
        ValueError: If the input parameters to the API are invalid.
        ConnectionError: If the server address is invalid or not reachable.
    """
    response = Client._post_data(
        relative_url=f"/model/train",
        json=dict(
            data_uuid=self.uuid,
            client_column=client_column,
            target_column=target_column,
            target=target,
            predict_after=int(predict_after.total_seconds()),
        ),
    )

    return Model(uuid=response["uuid"])

# %% ../../notebooks/API_DataSource.ipynb 41
add_example_to_docs(DataSource.train, _docstring_example.__doc__)  # type: ignore



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/CLI_Token.ipynb.

# %% auto 0
__all__ = ['logger']

# %% ../../notebooks/CLI_Token.ipynb 3
from typing import *

# %% ../../notebooks/CLI_Token.ipynb 4
import os
import typer
import logging
import time
from datetime import datetime, timedelta

from airt.client import Client, User
from airt.logger import get_logger, set_level
from airt.constant import (
    SERVER_URL,
    SERVICE_USERNAME,
    SERVICE_PASSWORD,
    CLIENT_NAME,
    SERVICE_TOKEN,
)
from airt.cli import helper

# %% ../../notebooks/CLI_Token.ipynb 6
SESSION_TIME_LIMIT = 10  # mins

# %% ../../notebooks/CLI_Token.ipynb 7
app = typer.Typer()

# %% ../../notebooks/CLI_Token.ipynb 9
logger = get_logger(__name__)

# %% ../../notebooks/CLI_Token.ipynb 12
@helper.requires_totp_or_otp(
    message_template_name="get_token", requires_auth_token=False
)
def token(
    username: Optional[str] = typer.Option(
        None,
        "--username",
        "-u",
        help="Username for the developer account. If None (default value), then the value from"
        f" **{SERVICE_USERNAME}** environment variable is used.",
    ),
    password: Optional[str] = typer.Option(
        None,
        "--password",
        "-p",
        help="Password for the developer account. If None (default value), then the value from"
        f" **{SERVICE_PASSWORD}** environment variable is used.",
    ),
    server: Optional[str] = typer.Option(
        None,
        "--server",
        "-s",
        help=f"The {CLIENT_NAME} server uri. If None (default value), then the value from **{SERVER_URL}** environment variable"
        " is used. If the variable is not set as well, then the default public server will be used. Please leave this"
        f" setting to default unless you are running the service in your own server (please email us to info@{CLIENT_NAME}.ai"
        " for that possibility).",
    ),
    otp: Optional[str] = typer.Option(
        None,
        "--otp",
        help=f"Dynamically generated six-digit verification code from the authenticator app or the OTP you have received via SMS."
        " Please do not pass this parameter if you haven't enabled the multi-factor authentication for your account.",
    ),
    sso_provider: Optional[str] = typer.Option(
        None,
        "--sso_provider",
        "-sp",
        help=f"Name of the Single sign-on (SSO) provider. At the moment, we only support google and github as SSO providers."
        " Please pass this parameter only if you have successfully enabled SSO for the provider.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output authentication token only.",
    ),
    debug: bool = typer.Option(
        False, "--debug", "-d", help="Set logger level to DEBUG and output everything."
    ),
) -> None:
    """Get application token for airt service from a username/password pair.

    To access the airt service, you must first create a developer account. To obtain one, please contact us at info@airt.ai.

    After successful verification, you will receive an email with the username and password for the developer account.

    Once you have the credentials, use them to get an access token by running **airt token** command. It is necessary to
    get an access token; otherwise, you won't be able to access all of the airt service's APIs. You can either pass the
    username, password, and server address as command line arguments or store them in the environment variables
    **AIRT_SERVICE_USERNAME**, **AIRT_SERVICE_PASSWORD**, and **AIRT_SERVER_URL**.

    If you've already enabled multi-factor authentication (MFA) for your account, you'll need to pass the dynamically
    generated six-digit verification code along with your username and password to generate new tokens.

    If the token is requested using Single sign-on (SSO), an authorization URL will be returned. Please copy and paste
    it into your preferred browser and complete the SSO provider authentication within 10 minutes. Otherwise,
    the SSO login will time out and you will need to re-request the token.

    Single sign-on (SSO) can be enabled for your account in three simple steps:

    1. Enable the SSO for a provider by calling the command `airt user sso enable` with the SSO provider name and an email address.
    At the moment, we only support "google" and "github" as SSO providers. We intend to support additional SSO providers in future releases.

    2. Before you can start generating new tokens with SSO, you must first authenticate with the SSO provider. Call the `airt token` command with
    the same SSO provider you have enabled in the step above to generate an SSO authorization URL. Please copy and paste it into your
    preferred browser and complete the authentication process with the SSO provider.

    3. After successfully authenticating with the SSO provider, an access token will be generated and returned. Please set it in the
    **AIRT_SERVICE_TOKEN** environment variable for accessing the airt service.
    """
    try:
        if debug:
            set_level(logging.DEBUG)
        else:
            set_level(logging.WARNING)

        if sso_provider is None:
            Client.get_token(
                username=username, password=password, server=server, otp=otp
            )

            if quiet:
                typer.echo(Client.auth_token)
            else:
                typer.echo(f"token: {Client.auth_token}")

        else:
            authorization_url = Client.get_token(
                username=username,
                password=password,
                server=server,
                otp=otp,
                sso_provider=sso_provider,
            )

            typer.echo(
                "\nPlease copy and paste the authorization URL below into your preferred browser and complete the SSO provider authentication "
                "within 10 minutes. Otherwise, the SSO login will time out and you will have to re-run the token command."
            )

            typer.echo(f"\n\n{authorization_url}\n")

            typer.echo(
                f"\nAfter successfully authenticating with the SSO provider, an access token will be returned. Please set it in the {SERVICE_TOKEN} "
                f"environment variable for accessing the `{CLIENT_NAME}` service."
            )

            typer.echo(
                "\nIf there are any errors, an error message will be displayed in the terminal and this command will be terminated.\n"
            )

            end_time = datetime.utcnow() + timedelta(minutes=SESSION_TIME_LIMIT)
            while datetime.utcnow() < end_time:
                err = None
                try:
                    Client.set_sso_token()
                    typer.echo(
                        f"\nSSO authentication is successful, please set the below token in the `{SERVICE_TOKEN}` environment variable for accessing the `{CLIENT_NAME}` service.\n"
                    )
                    typer.echo(f"{Client.auth_token}\n")
                    break
                except ValueError as e:
                    err = str(e)
                    if "SSO authentication is not complete" not in str(e):
                        raise ValueError(f"\n{e}")
                time.sleep(1)
            if err is not None:
                raise ValueError(
                    f"\nYour session has expired. Please call the {CLIENT_NAME} token command with the sso provider and try again."
                )

        if not quiet:
            details = User.details()
            status = helper.get_phone_registration_status(details)
            if status is not None:
                typer.echo(status)

    except KeyError as e:
        typer.echo(message=f"Error: {e}", err=True)
        typer.echo(f"\nTry '{CLIENT_NAME} token --help' for help.")
        raise typer.Exit(code=1)

    except Exception as e:
        typer.echo(message=f"Error: {e}", err=True)
        if ("Invalid OTP" in str(e)) or ("OTP is required" in str(e)):
            raise ValueError(e)
        raise typer.Exit(code=1)

# %% ../../notebooks/CLI_Token.ipynb 13
app.command()(token)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/API_DataBlob.ipynb.

# %% auto 0
__all__ = ['DataBlob']

# %% ../../notebooks/API_DataBlob.ipynb 4
from typing import *

# %% ../../notebooks/API_DataBlob.ipynb 5
import os
import requests
from contextlib import contextmanager

import pandas as pd
from fastcore.foundation import patch
from datetime import datetime, timedelta
from pathlib import Path
from tqdm import tqdm

from airt.logger import get_logger, set_level
from airt.helper import (
    get_data,
    post_data,
    delete_data,
    add_ready_column,
    generate_df,
    get_values_from_item,
    get_attributes_from_instances,
    add_example_to_docs,
)

from airt.components.client import Client
from airt.components.datasource import DataSource
from airt.components.progress_status import ProgressStatus
from airt.constant import CLIENT_DB_USERNAME, CLIENT_DB_PASSWORD

# %% ../../notebooks/API_DataBlob.ipynb 7
logger = get_logger(__name__)

# %% ../../notebooks/API_DataBlob.ipynb 10
DEFAULT_AZURE_BLOB_STORAGE_REGION = "westeurope"
DEFAULT_S3_REGION = "eu-west-1"

# %% ../../notebooks/API_DataBlob.ipynb 11
class DataBlob:
    """A class for importing and processing data from sources such as CSV/parquet files, databases, AWS S3 buckets, and Azure Blob Storage.

    Currently, the only way to instantiate the DataBlob class is to call one of the following static methods
    `from_local`, `from_mysql`, `from_clickhouse`, `from_s3`, or `from_azure_blob_storage` which imports the data in
    the parquet file format from:

     - a local CSV/parquet file,

     - a MySql database,

     - a ClickHouse database

     - an AWS S3 bucket, and

     - an Azure Blob Storage respectively.

    We intend to support additional databases and storage mediums in future releases.
    """

    BASIC_DB_COLS = [
        "uuid",
        "datasources",
        "type",
        "source",
        "region",
        "cloud_provider",
        "tags",
        "pulled_on",
        "completed_steps",
        "total_steps",
        "folder_size",
    ]

    ALL_DB_COLS = BASIC_DB_COLS + ["user", "error", "disabled"]

    COLS_TO_RENAME = {
        "uuid": "datablob_uuid",
        "datasources": "datasource_uuids",
        "user": "user_uuid",
    }

    _default_provider_and_regions: List[Tuple[str, str]] = []

    def __init__(
        self,
        uuid: str,
        type: Optional[str] = None,
        source: Optional[str] = None,
        region: Optional[str] = None,
        cloud_provider: Optional[str] = None,
        datasources: Optional[List[str]] = None,
        total_steps: Optional[int] = None,
        completed_steps: Optional[int] = None,
        folder_size: Optional[int] = None,
        disabled: Optional[bool] = None,
        pulled_on: Optional[str] = None,
        user: Optional[str] = None,
        tags: Optional[List] = None,
        error: Optional[str] = None,
    ):
        """Constructs a new DataBlob instance.

        Warning:
            Do not construct this object directly by calling the constructor, please use `from_s3`, `from_azure_blob_storage`,
            `from_mysql`, `from_clickhouse` or `from_local` methods instead.

        Args:
            uuid: Datablob uuid.
            source: The URI of the data that was used to create the datablob.
            type: The type of source used to generate the datablob. Depending on the source type, one of the following
                values will be assigned: "s3", "local", "db", or "azure_blob_storage".
            region: The destination cloud provider's region to store the datablob. If None (default value) then the default region will be assigned based on the cloud provider.
            cloud_provider: Cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
            datasources: The uuids of the datasources created from the datablob.
            total_steps: The number of steps required to upload the datablob to the server.
            completed_steps: The number of steps completed during the datablob's upload to the server.
            folder_size: The uploaded datablob's size in bytes.
            disabled: A flag that indicates the datablob's status. If the datablob is deleted, then **False** will be set.
            pulled_on: The most recent date the datablob was uploaded.
            user: The uuid of the user who created the datablob.
            tags: Tag names associated with the datablob.
            error: Contains the error message if the processing of the datablob fails.
        """
        self.uuid = uuid
        self.type = type
        self.source = source
        self.region = region
        self.cloud_provider = cloud_provider
        self.datasources = datasources
        self.total_steps = total_steps
        self.completed_steps = completed_steps
        self.folder_size = folder_size
        self.disabled = disabled
        self.pulled_on = pulled_on
        self.user = user
        self.tags = tags
        self.error = error

    @staticmethod
    def _get_tag_name_and_datasource_id(res: Dict[str, Any]) -> Dict[str, Any]:
        """Get tag name and datasource ids as string seperated by comma.

        Args:
            res: The response object.

        Returns:
            The modified response object with tag name and datasource ids as string seperated by comma.
        """
        res["tags"] = get_values_from_item(res["tags"], "name")
        res["datasources"] = get_values_from_item(res["datasources"])

        return res

    @staticmethod
    def from_s3(
        *,
        uri: str,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        cloud_provider: Optional[str] = None,
        region: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> "DataBlob":
        """Create and return a datablob that encapsulates the data from an AWS S3 bucket.

        Args:
            uri: AWS S3 bucket uri.
            access_key: Access key for the S3 bucket. If **None** (default value), then the value
                from **AWS_ACCESS_KEY_ID** environment variable will be used.
            secret_key: Secret key for the S3 bucket. If **None** (default value), then the value
                from **AWS_SECRET_ACCESS_KEY** environment variable will be used.
            cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
                If **None** (default value), then **aws**  will be used as the cloud storage provider.
            region: The region of the destination cloud provider where the datablob will be stored. If **None** (default value) then the default region will be assigned based on
                the cloud provider. In the case of **aws**, the datablob's source bucket region will be used, whereas **azure** will use **westeurope**. The supported AWS regions
                are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1, ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1,
                us-east-1, us-east-2, us-west-1, us-west-2. The supported Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast,
                brazilsouth, canadacentral, canadaeast, centralindia, centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast,
                japanwest, koreacentral, koreasouth, northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia,
                switzerlandnorth, switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.
            tag: A string to tag the datablob. If not passed, then the tag **latest** will be assigned to the datablob.

        Returns:
            An instance of the `DataBlob` class.

        Raises:
            ValueError: If parameters to the API are invalid.
            ConnectionError: If the server address is invalid or not reachable.

        Here's an example of how to create a Datablob from an AWS S3 bucket:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the access_key and the secret_key are set in the
            # AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment variables. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_s3(
                uri="{fill in uri}",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Display the status in a progress bar
            db.progress_bar()

            # Print the details of the newly created datablob
            # If the upload is successful, the ready flag should be set to True
            print(db.details())
            ```
        """
        access_key = (
            access_key if access_key is not None else os.environ["AWS_ACCESS_KEY_ID"]
        )
        secret_key = (
            secret_key
            if secret_key is not None
            else os.environ["AWS_SECRET_ACCESS_KEY"]
        )

        cloud_provider, region = DataBlob._get_cloud_provider_and_region(cloud_provider=cloud_provider, region=region, set_source_region=True)  # type: ignore

        response = Client._post_data(
            relative_url="/datablob/from_s3",
            json=dict(
                uri=uri,
                access_key=access_key,
                secret_key=secret_key,
                region=region,
                cloud_provider=cloud_provider,
                tag=tag,
            ),
        )

        return DataBlob(
            uuid=response["uuid"], type=response["type"], source=response["source"]
        )

    @classmethod
    def from_azure_blob_storage(
        cls,
        uri: str,
        credential: str,
        cloud_provider: Optional[str] = None,
        region: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> "DataBlob":
        """Create and return a datablob that encapsulates the data from an Azure Blob Storage.

        Args:
            uri: Azure Blob Storage URI of the source file.
            credential: Credential to access the Azure Blob Storage.
            cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
                If **None** (default value), then **azure**  will be used as the cloud storage provider.
            region: The destination cloud provider's region to store the datablob. If **None** (default value) then the default region will be assigned based on the cloud
                provider. In the case of **aws**, **eu-west-1** will be used and in the case of **azure**, **westeurope** will be used. The supported AWS regions
                are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1, ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1,
                us-east-1, us-east-2, us-west-1, us-west-2. The supported Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast,
                brazilsouth, canadacentral, canadaeast, centralindia, centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast,
                japanwest, koreacentral, koreasouth, northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia,
                switzerlandnorth, switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.
            tag: A string to tag the datablob. If not passed, then the tag **latest** will be assigned to the datablob.

        Returns:
            An instance of the `DataBlob` class.

        Raises:
            ValueError: If parameters to the API are invalid.
            ConnectionError: If the server address is invalid or not reachable.

        To create a Datablob from Azure Blob Storage, you must have a valid Azure Blob Storage credential.

        If you don't know how to get the Azure Blob Storage credential, you can follow the below python example. It's one of the ways to get the Azure Blob Storage credential.

        - If you don't already have it, please install the Azure Storage Management (azure-mgmt-storage) and Azure Resource Management (azure-mgmt-resource) python client libraries using pip.

        - Ensure the following four environment variables are set into your current working environment with appropriate values.

            - AZURE_TENANT_ID

            - AZURE_CLIENT_ID

            - AZURE_CLIENT_SECRET

            - AZURE_SUBSCRIPTION_ID

        - Assign the resource group name in the GROUP_NAME variable and the storage account name in the STORAGE_ACCOUNT_NAME variable.

        - Below is a sample code to create a datablob and storing it in S3. Please copy it and replace the placeholders with appropriate values

        Example:
            ```python
            # Importing necessary libraries
            import os

            from azure.identity import DefaultAzureCredential
            from azure.mgmt.storage import StorageManagementClient

            from  airt.client import Client, DataBlob

            # Create a credential for accessing Azure Blob Storage
            # Setting the required environment variables
            os.environ["AZURE_SUBSCRIPTION_ID"] = "{fill in azure_subscription_id}"
            os.environ["AZURE_CLIENT_ID"] = "{fill in azure_client_id}"
            os.environ["AZURE_CLIENT_SECRET"] = "{fill in azure_client_secret}"
            os.environ["AZURE_TENANT_ID"]= "{fill in azure_tenant_id}"

            # Setting the resource group name and storage account name
            azure_group_name = "{fill in azure_group_name}"
            azure_storage_account_name = "{fill in azure_storage_account_name}"

            # Retrieving the credential
            azure_storage_client = StorageManagementClient(
                DefaultAzureCredential(), os.environ["AZURE_SUBSCRIPTION_ID"]
            )
            azure_storage_keys = azure_storage_client.storage_accounts.list_keys(
                azure_group_name, azure_storage_account_name
            )
            azure_storage_keys = {v.key_name: v.value for v in azure_storage_keys.keys}
            credential = azure_storage_keys['key1']


            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-1 (default), feel free to change the cloud provider and
            # the region to suit your needs.
            db = DataBlob.from_azure_blob_storage(
                uri="{fill in uri}",
                cloud_provider="aws",
                credential=credential
            )

            # Display the status in a progress bar
            db.progress_bar()

            # Print the details of the newly created datablob
            # If the upload is successful, the ready flag should be set to True
            print(db.details())
            ```
        """
        cloud_provider, region = DataBlob._get_cloud_provider_and_region(cloud_provider=cloud_provider, region=region, default_cloud_provider="azure")  # type: ignore

        response = Client._post_data(
            relative_url="/datablob/from_azure_blob_storage",
            json=dict(
                uri=uri,
                credential=credential,
                region=region,
                cloud_provider=cloud_provider,
                tag=tag,
            ),
        )

        return DataBlob(
            uuid=response["uuid"], type=response["type"], source=response["source"]
        )

    @staticmethod
    def from_mysql(
        *,
        host: str,
        database: str,
        table: str,
        port: int = 3306,
        cloud_provider: Optional[str] = None,
        region: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        tag: Optional[str] = None,
    ) -> "DataBlob":
        """Create and return a datablob that encapsulates the data from a mysql database.

        If the database requires authentication, pass the username/password as parameters or store it in
        the **AIRT_CLIENT_DB_USERNAME** and **AIRT_CLIENT_DB_PASSWORD** environment variables.

        Args:
            host: Remote database host name.
            database: Database name.
            table: Table name.
            port: Host port number. If not passed, then the default value **3306** will be used.
            cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
                If **None** (default value), then **aws**  will be used as the cloud storage provider.
            region: The destination cloud provider's region to store the datablob. If **None** (default value) then the default region will be assigned based on the cloud
                provider. In the case of **aws**, **eu-west-1** will be used and in the case of **azure**, **westeurope** will be used. The supported AWS regions
                are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1, ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1,
                us-east-1, us-east-2, us-west-1, us-west-2. The supported Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast,
                brazilsouth, canadacentral, canadaeast, centralindia, centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast,
                japanwest, koreacentral, koreasouth, northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia,
                switzerlandnorth, switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.
            username: Database username. If not passed, the default value **"root"** will be used unless the value is explicitly set in the environment variable
                **AIRT_CLIENT_DB_USERNAME**.
            password: Database password. If not passed, the default value **""** will be used unless the value is explicitly set in the environment variable
                **AIRT_CLIENT_DB_PASSWORD**.
            tag: A string to tag the datablob. If not passed, then the tag **latest** will be assigned to the datablob.

        Returns:
           An instance of the `DataBlob` class.

        Raises:
            ValueError: If parameters to the API are invalid.
            ConnectionError: If the server address is invalid or not reachable.

        Here's an example of how to create a Datablob from a MySQL database:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_mysql(
                username="{fill in database_username}",
                password="{fill in database_password}",
                host="{fill in host}",
                database="{fill in database}",
                table="{fill in table}",
                port="{fill in port}",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Display the status in a progress bar
            db.progress_bar()

            # Print the details of the newly created datablob
            # If the upload is successful, the ready flag should be set to True
            print(db.details())
            ```
        """
        username = (
            username
            if username is not None
            else os.environ.get(CLIENT_DB_USERNAME, "root")
        )

        password = (
            password if password is not None else os.environ.get(CLIENT_DB_PASSWORD, "")
        )

        cloud_provider, region = DataBlob._get_cloud_provider_and_region(cloud_provider, region)  # type: ignore

        json_req = dict(
            host=host,
            port=port,
            username=username,
            password=password,
            database=database,
            table=table,
            region=region,
            cloud_provider=cloud_provider,
            tag=tag,
        )

        response = Client._post_data(
            relative_url=f"/datablob/from_mysql", json=json_req
        )

        return DataBlob(
            uuid=response["uuid"], type=response["type"], source=response["source"]
        )

    @staticmethod
    def from_clickhouse(
        *,
        host: str,
        database: str,
        table: str,
        protocol: str,
        index_column: str,
        timestamp_column: str,
        port: int = 0,
        cloud_provider: Optional[str] = None,
        region: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        filters: Optional[Dict[str, Any]] = None,
        tag: Optional[str] = None,
    ) -> "DataBlob":
        """Create and return a datablob that encapsulates the data from a ClickHouse database.

        If the database requires authentication, pass the username/password as parameters or store it in
        the **CLICKHOUSE_USERNAME** and **CLICKHOUSE_PASSWORD** environment variables.

        Args:
            host: Remote database host name.
            database: Database name.
            table: Table name.
            protocol: Protocol to use. The valid values are "native" and "http".
            index_column: The column to use as index (row labels).
            timestamp_column: Timestamp column name in the tabel.
            port: Host port number. If not passed, then the default value **0** will be used.
            cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
                If **None** (default value), then **aws**  will be used as the cloud storage provider.
            region: The destination cloud provider's region to store the datablob. If **None** (default value) then the default region will be assigned based on the cloud
                provider. In the case of **aws**, **eu-west-1** will be used and in the case of **azure**, **westeurope** will be used. The supported AWS regions
                are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1, ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1,
                us-east-1, us-east-2, us-west-1, us-west-2. The supported Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast,
                brazilsouth, canadacentral, canadaeast, centralindia, centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast,
                japanwest, koreacentral, koreasouth, northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia,
                switzerlandnorth, switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.
            username: Database username. If not passed, the default value "root" will be used unless the value is explicitly set in the environment variable
                **CLICKHOUSE_USERNAME**.
            password: Database password. If not passed, the default value "root" will be used unless the value is explicitly set in the environment variable
                **CLICKHOUSE_PASSWORD**.
            filters: Additional parameters to be used when importing data. For example, if you want to filter and extract data only for a specific user_id, pass {"user_id": 1}.
            tag: A string to tag the datablob. If not passed, then the tag **latest** will be assigned to the datablob.

        Returns:
           An instance of the `DataBlob` class.

        Raises:
            ValueError: If parameters to the API are invalid.
            ConnectionError: If the server address is invalid or not reachable.

        Here's an example of how to create a Datablob from a ClickHouse database:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_clickhouse(
                username="{fill in database_username}",
                password="{fill in database_password}",
                host="{fill in host}",
                database="{fill in database}",
                table="{fill in table}",
                index_column="{fill in index_column}",
                timestamp_column="{fill in timestamp_column}",
                port="{fill in port}",
                filters={fill in filters},
                protocol="native",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Display the status in a progress bar
            db.progress_bar()

            # Print the details of the newly created datablob
            # If the upload is successful, the ready flag should be set to True
            print(db.details())
            ```
        """
        username = (
            username
            if username is not None
            else os.environ.get("CLICKHOUSE_USERNAME", "root")
        )

        password = (
            password
            if password is not None
            else os.environ.get("CLICKHOUSE_PASSWORD", "")
        )

        cloud_provider, region = DataBlob._get_cloud_provider_and_region(cloud_provider, region)  # type: ignore

        json_req = dict(
            host=host,
            database=database,
            table=table,
            protocol=protocol,
            port=port,
            username=username,
            password=password,
            index_column=index_column,
            timestamp_column=timestamp_column,
            filters=filters,
            region=region,
            cloud_provider=cloud_provider,
            tag=tag,
        )

        response = Client._post_data(
            relative_url=f"/datablob/from_clickhouse", json=json_req
        )

        return DataBlob(
            uuid=response["uuid"], type=response["type"], source=response["source"]
        )

    @staticmethod
    def _upload_to_s3_with_retry(
        file_to_upload: Path,
        presigned_url: str,
        presigned_fields: Dict[str, Any],
        max_retry: int = 3,
        curr_iteration: int = 1,
    ):
        """Upload local files to s3 using presigned url

        Args:
            file_to_upload: path of file to upload
            presigned_url: presigned url to upload to
            presigned_fields: presigned fields provided by boto3
            max_retry: maximum retry count
            curr_iteration: current iteration count for internal use
        """
        try:
            with open(file_to_upload, "rb") as f:
                files = {"file": (str(file_to_upload), f)}
                response = requests.post(
                    presigned_url, data=presigned_fields, files=files
                )
                if not response.status_code == 204:
                    raise ValueError(response.text)

        except requests.exceptions.ConnectionError as e:
            if curr_iteration == max_retry:
                raise e
            DataBlob._upload_to_s3_with_retry(
                file_to_upload,
                presigned_url,
                presigned_fields,
                max_retry,
                curr_iteration + 1,
            )

    @staticmethod
    def from_local(
        path: Union[str, Path],
        cloud_provider: Optional[str] = None,
        region: Optional[str] = None,
        tag: Optional[str] = None,
        show_progress: Optional[bool] = True,
    ) -> "DataBlob":
        """Create and return a datablob from local file.

        The API currently allows users to create datablobs from CSV or Parquet files. We intend to support additional file formats in future releases.

        Args:
            path: The relative or absolute path to a local file or to a directory containing the source files.
            cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
                If **None** (default value), then **aws**  will be used as the cloud storage provider.
            region: The destination cloud provider's region to store the datablob. If **None** (default value) then the default region will be assigned based on the cloud
                provider. In the case of **aws**, **eu-west-1** will be used and in the case of **azure**, **westeurope** will be used. The supported AWS regions
                are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1, ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1,
                us-east-1, us-east-2, us-west-1, us-west-2. The supported Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast,
                brazilsouth, canadacentral, canadaeast, centralindia, centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast,
                japanwest, koreacentral, koreasouth, northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia,
                switzerlandnorth, switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.
            tag: A string to tag the datablob. If not passed, then the tag **latest** will be assigned to the datablob.
            show_progress: Flag to set the progressbar visibility. If not passed, then the default value **True** will be used.

        Returns:
           An instance of the `DataBlob` class.

        Raises:
            ValueError: If parameters to the API are invalid.
            ConnectionError: If the server address is invalid or not reachable.

        Here's an example of how to create a Datablob from a local file:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_local(
                path="{fill in path}",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Display the status in a progress bar
            db.progress_bar()

            # Print the details of the newly created datablob
            # If the upload is successful, the ready flag should be set to True
            print(db.details())

            ```
        """
        path = Path(path)
        cloud_provider, region = DataBlob._get_cloud_provider_and_region(cloud_provider, region)  # type: ignore

        # Step 1: get presigned URL
        _path = f"local:{str(path)}"

        response = Client._post_data(
            relative_url=f"/datablob/from_local/start",
            json=dict(
                path=_path, region=region, cloud_provider=cloud_provider, tag=tag
            ),
        )

        # Step 2: download the csv to the s3 bucket
        files = list(path.glob("*")) if path.is_dir() else [path]

        # Initiate progress bar
        t = tqdm(total=len(files), disable=not show_progress)

        for file_to_upload in files:
            DataBlob._upload_to_s3_with_retry(
                file_to_upload=file_to_upload,
                presigned_url=response["presigned"]["url"],
                presigned_fields=response["presigned"]["fields"],
            )
            t.update()

        t.close()
        return DataBlob(uuid=response["uuid"], type=response["type"])

    @staticmethod
    def ls(
        offset: int = 0,
        limit: int = 100,
        disabled: bool = False,
        completed: bool = False,
    ) -> List["DataBlob"]:
        """Return the list of DataBlob instances

        Args:
            offset: The number of datablobs to offset at the beginning. If **None**,
                then the default value **0** will be used.
            limit: The maximum number of datablobs to return from the server. If **None**,
                then the default value **100** will be used.
            disabled: If set to **True**, then only the deleted datablobs will be returned.
                Else, the default value **False** will be used to return only the list
                of active datablobs.
            completed: If set to **True**, then only the datablobs that are successfully downloaded
                to the server will be returned. Else, the default value **False** will be used to
                return all the datablobs.

        Returns:
            A list of DataBlob instances available in the server.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
        """
        lists = Client._get_data(
            relative_url=f"/datablob/?disabled={disabled}&completed={completed}&offset={offset}&limit={limit}"
        )

        dbx = [
            DataBlob(
                uuid=db["uuid"],
                type=db["type"],
                source=db["source"],
                region=db["region"],
                cloud_provider=db["cloud_provider"],
                datasources=db["datasources"],
                total_steps=db["total_steps"],
                completed_steps=db["completed_steps"],
                folder_size=db["folder_size"],
                disabled=db["disabled"],
                pulled_on=db["pulled_on"],
                user=db["user"],
                tags=db["tags"],
                error=db["error"],
            )
            for db in lists
        ]

        return dbx

    @staticmethod
    def as_df(dbx: List["DataBlob"]) -> pd.DataFrame:
        """Return the details of datablob instances as a pandas dataframe.

        Args:
            dbx: List of datablob instances.

        Returns:
            Details of all the datablobs in a dataframe.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
        """
        db_lists = get_attributes_from_instances(dbx, DataBlob.ALL_DB_COLS)  # type: ignore

        for db in db_lists:
            db = DataBlob._get_tag_name_and_datasource_id(db)

        lists_df = generate_df(db_lists, DataBlob.BASIC_DB_COLS)
        df = add_ready_column(lists_df)

        df = df.rename(columns=DataBlob.COLS_TO_RENAME)

        return df

    def is_ready(self) -> bool:
        """Check if the method's progress is complete.

        !!! info

            This method will return `True` immediately and will not wait for the progress to finish
            if the datablob is created using the `from_local` method.

        Returns:
            **True** if the upload progress is completed, else **False**.
        """
        if self.type in ["local"]:
            return True

        progress_status = ProgressStatus(relative_url=f"/datablob/{self.uuid}")

        return progress_status.is_ready()

    def progress_bar(self, sleep_for: Union[int, float] = 5, timeout: int = 0):
        """Blocks the execution and displays a progress bar showing the remote action progress.

        !!! info

            This method will not check the progress if the datablob is created using the
            `from_local` method.

        Args:
            sleep_for: The time interval in seconds between successive API calls.
            timeout: The maximum time allowed in seconds for the asynchronous call to complete. If not the
                progressbar will be terminated.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
            TimeoutError: in case of connection timeout.
        """
        if self.type not in ["local"]:
            progress_status = ProgressStatus(
                relative_url=f"/datablob/{self.uuid}",
                sleep_for=sleep_for,
                timeout=timeout,
            )

            progress_status.progress_bar()

    def wait(self, sleep_for: Union[int, float] = 1, timeout: int = 0):
        """Blocks execution while waiting for the remote action to complete.

        !!! info

            This method will not check the progress if the datablob is created using the
            `from_local` method.

        Args:
            sleep_for: The time interval in seconds between successive API calls.
            timeout: The maximum time allowed in seconds for the asynchronous call to complete. If not the
                progressbar will be terminated.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
            TimeoutError: in case of timeout.
        """
        if self.type not in ["local"]:
            progress_status = ProgressStatus(
                relative_url=f"/datablob/{self.uuid}",
                sleep_for=sleep_for,
                timeout=timeout,
            )

            progress_status.wait()

    def to_datasource(
        self,
        *,
        file_type: str,
        index_column: str,
        sort_by: Union[str, List[str]],
        deduplicate_data: bool = False,
        blocksize: str = "256MB",
        **kwargs,
    ) -> DataSource:
        raise NotImplementedError()

    def details(self) -> pd.DataFrame:
        raise NotImplementedError()

    def tag(self, name: str) -> pd.DataFrame:
        raise NotImplementedError()

    def delete(self) -> pd.DataFrame:
        raise NotImplementedError()

# %% ../../notebooks/API_DataBlob.ipynb 12
def _docstring_example():
    """
    Example:
        ```python
        # Importing necessary libraries
        from  airt.client import Client, DataBlob

        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a datablob
        # In this example, the datablob will be stored in an AWS S3 bucket. The
        # access_key and the secret_key are set in the AWS_ACCESS_KEY_ID and
        # AWS_SECRET_ACCESS_KEY environment variables, and the region is set to
        # eu-west-3; feel free to change the cloud provider and the region to
        # suit your needs.
        db = DataBlob.from_s3(
            uri="{fill in uri}",
            cloud_provider="aws",
            region="eu-west-3"
        )

        # Display the status in a progress bar
        # Call the wait method to wait for the progress to finish but
        # without displaying an interactive progress bar.
        db.progress_bar()

        # Display the ready status
        # If the datablob is successfully uploaded, True will be returned.
        print(db.is_ready())

        # Print the details of the newly created datablob
        print(db.details())

        # Display the details of all datablob created by the currently
        # logged-in user
        print(DataBlob.as_df(DataBlob.ls()))

        # Create a datasource
        ds = db.to_datasource(
            file_type="{fill in file_type}",
            index_column="{fill in index_column}",
            sort_by="{fill in sort_by}",
        )

        # Display the status in a progress bar
        ds.progress_bar()

        # Display the head of the data to ensure everything is fine.
        print(ds.head())

        # Tag the datablob
        print(db.tag(name="{fill in tag_name}"))

        # Delete the datablob
        print(db.delete())
        ```
    """
    pass

# %% ../../notebooks/API_DataBlob.ipynb 14
add_example_to_docs(DataBlob, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataBlob.ls, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataBlob.as_df, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataBlob.wait, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataBlob.is_ready, _docstring_example.__doc__)  # type: ignore
add_example_to_docs(DataBlob.progress_bar, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataBlob.ipynb 15
@patch(cls_method=True)
@contextmanager
def set_default_cloud_provider(
    cls: DataBlob, cloud_provider: str, region: Optional[str] = None
) -> Iterator[None]:
    """Sets the default destination value for the cloud_provider and the region.

    Whenever you call the from_\* methods of the `DataBlob` class inside this context manager, the destination cloud_provider and region set in this context
    will be passed to the from_\* methods, unless you explicitely override it in the parameter.

    Args:
        cloud_provider: The destination cloud storage provider's name to store the datablob. Currently, the API only supports **aws** and **azure** as cloud storage providers.
        region: The destination cloud provider's region to store the datablob. The supported AWS regions are: ap-northeast-1, ap-northeast-2, ap-south-1, ap-southeast-1,
            ap-southeast-2, ca-central-1, eu-central-1, eu-north-1, eu-west-1, eu-west-2, eu-west-3, sa-east-1, us-east-1, us-east-2, us-west-1, us-west-2. The supported
            Azure Blob Storage regions are: australiacentral, australiacentral2, australiaeast, australiasoutheast, brazilsouth, canadacentral, canadaeast, centralindia,
            centralus, eastasia, eastus, eastus2, francecentral, francesouth, germanynorth, germanywestcentral, japaneast, japanwest, koreacentral, koreasouth,
            northcentralus, northeurope, norwayeast, norwaywest, southafricanorth, southafricawest, southcentralus, southeastasia, southindia, switzerlandnorth,
            switzerlandwest, uaecentral, uaenorth, uksouth, ukwest, westcentralus, westeurope, westindia, westus, westus2.

    Returns:
        A context manager that specifies the cloud provider and region to use.

    Here's an example of creating a datablob from Azure Blob Storage and storing it in AWS S3:

    Example:
        ```python
        # Importing necessary libraries
        import os

        from azure.identity import DefaultAzureCredential
        from azure.mgmt.storage import StorageManagementClient

        from  airt.client import Client, DataBlob

        # Create a credential for accessing Azure Blob Storage
        # Setting the required environment variables
        os.environ["AZURE_SUBSCRIPTION_ID"] = "{fill in azure_subscription_id}"
        os.environ["AZURE_CLIENT_ID"] = "{fill in azure_client_id}"
        os.environ["AZURE_CLIENT_SECRET"] = "{fill in azure_client_secret}"
        os.environ["AZURE_TENANT_ID"]= "{fill in azure_tenant_id}"

        # Setting the resource group name and storage account name
        azure_group_name = "{fill in azure_group_name}"
        azure_storage_account_name = "{fill in azure_storage_account_name}"

        # Retrieving the credential
        azure_storage_client = StorageManagementClient(
            DefaultAzureCredential(), os.environ["AZURE_SUBSCRIPTION_ID"]
        )
        azure_storage_keys = azure_storage_client.storage_accounts.list_keys(
            azure_group_name, azure_storage_account_name
        )
        azure_storage_keys = {v.key_name: v.value for v in azure_storage_keys.keys}
        credential = azure_storage_keys['key1']


        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a datablob
        # In this example, the datablobs created inside the context manager will be
        # stored in an AWS S3 bucket with the region set to eu-west-3.
        with DataBlob.set_default_cloud_provider(
            cloud_provider="aws",
            region="eu-west-3"
        ):
            db = DataBlob.from_azure_blob_storage(
                uri="{fill in uri}",
                credential=credential
            )

        # Display the status in a progress bar
        db.progress_bar()

        # Print the details of the newly created datablob
        # If the upload is successful, the ready flag should be set to True
        print(db.details())
        ```
    """

    cls._default_provider_and_regions.append((cloud_provider, region))  # type: ignore

    yield

    cls._default_provider_and_regions.pop()


@patch(cls_method=True)
def _get_default_provider_and_regions(
    cls: DataBlob,
) -> Tuple[Optional[str], Optional[str]]:

    if len(cls._default_provider_and_regions) == 0:
        return None, None

    else:
        return cls._default_provider_and_regions[-1]


@patch(cls_method=True)
def _get_cloud_provider_and_region(
    cls: DataBlob,
    cloud_provider: Optional[str] = None,
    region: Optional[str] = None,
    set_source_region: Optional[bool] = False,
    default_cloud_provider: str = "aws",
) -> Tuple[str, Optional[str]]:

    if (cloud_provider is None) and (region is not None):
        raise ValueError(
            "You must specify a cloud_provider if are specifying a region."
        )

    if (cloud_provider is None) and (region is None):
        cloud_provider, region = cls._get_default_provider_and_regions()  # type: ignore

    if cloud_provider is None:
        ret_val_cloud_provider = default_cloud_provider
    else:
        ret_val_cloud_provider = cloud_provider

    if region is None:
        if ret_val_cloud_provider == "azure":
            region = DEFAULT_AZURE_BLOB_STORAGE_REGION
        else:
            region = None if set_source_region else DEFAULT_S3_REGION

    return ret_val_cloud_provider, region

# %% ../../notebooks/API_DataBlob.ipynb 27
@patch
def details(self: DataBlob) -> pd.DataFrame:
    """Return details of a datablob.

    Returns:
        The datablob details as a pandas dataframe.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """

    details = Client._get_data(relative_url=f"/datablob/{self.uuid}")

    details = DataBlob._get_tag_name_and_datasource_id(details)

    details_df = pd.DataFrame([details])[DataBlob.ALL_DB_COLS]

    details_df = details_df.rename(columns=DataBlob.COLS_TO_RENAME)

    return add_ready_column(details_df)

# %% ../../notebooks/API_DataBlob.ipynb 28
add_example_to_docs(DataBlob.details, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataBlob.ipynb 29
DataBlob.details.__doc__ = DataBlob.details.__doc__ + f"\n    Columns in the resulting dataframe are: {', '.join(DataBlob.ALL_DB_COLS)}."  # type: ignore

# %% ../../notebooks/API_DataBlob.ipynb 51
@patch
def to_datasource(
    self: DataBlob,
    *,
    file_type: str,
    index_column: str,
    sort_by: Union[str, List[str]],
    deduplicate_data: bool = False,
    blocksize: str = "256MB",
    **kwargs,
) -> DataSource:
    """Process the datablob and return a datasource object.

    Args:
        file_type: The file type of the datablob. Currently, the API only supports **"csv"** and **"parquet"** as file types.
        index_column: The column to use as index (row labels).
        sort_by: The column(s) to sort the data. Can either be a string or a list of strings.
        deduplicate_data: If set to **True** (default value **False**), the datasource will be created with duplicate rows removed.
        blocksize: The number of bytes used to split larger files. If None, then the default value **256MB** will be used.
        kwargs: Additional keyword arguments to use while processing the data.e.g: To skip 100 lines from the bottom of file,
            pass **{"skipfooter": 100}

    Returns:
        An instance of the `DataSource` class.

    Raises:
        ValueError: If the CSV file processing fails.
        ConnectionError: If the server address is invalid or not reachable.
    """
    json_req = dict(
        file_type=file_type,
        deduplicate_data=deduplicate_data,
        index_column=index_column,
        sort_by=sort_by,
        blocksize=blocksize,
        kwargs=kwargs,
    )
    response = Client._post_data(
        relative_url=f"/datablob/{self.uuid}/to_datasource", json=json_req
    )

    return DataSource(uuid=response["uuid"])

# %% ../../notebooks/API_DataBlob.ipynb 52
add_example_to_docs(DataBlob.to_datasource, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataBlob.ipynb 55
@patch
def tag(self: DataBlob, name: str) -> pd.DataFrame:
    """Tag an existing datablob in the server.

    Args:
        name: A string to tag the datablob.

    Returns:
        A pandas dataframe with the details of the tagged datablob.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """
    response = Client._post_data(
        relative_url=f"/datablob/{self.uuid}/tag", json=dict(name=name)
    )

    response = DataBlob._get_tag_name_and_datasource_id(response)

    df = pd.DataFrame([response])[DataBlob.BASIC_DB_COLS]

    df = df.rename(columns=DataBlob.COLS_TO_RENAME)

    return add_ready_column(df)

# %% ../../notebooks/API_DataBlob.ipynb 56
add_example_to_docs(DataBlob.tag, _docstring_example.__doc__)  # type: ignore

# %% ../../notebooks/API_DataBlob.ipynb 58
@patch
def delete(self: DataBlob) -> pd.DataFrame:
    """Delete a datablob from the server.

    Returns:
        A pandas DataFrame encapsulating the details of the deleted datablob.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
    """

    response = Client._delete_data(relative_url=f"/datablob/{self.uuid}")

    response = DataBlob._get_tag_name_and_datasource_id(response)

    df = pd.DataFrame([response])[DataBlob.BASIC_DB_COLS]

    df = df.rename(columns=DataBlob.COLS_TO_RENAME)

    return add_ready_column(df)

# %% ../../notebooks/API_DataBlob.ipynb 59
add_example_to_docs(DataBlob.delete, _docstring_example.__doc__)  # type: ignore



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/CLI_Pred.ipynb.

# %% auto 0
__all__ = ['logger']

# %% ../../notebooks/CLI_Pred.ipynb 3
from typing import *

# %% ../../notebooks/CLI_Pred.ipynb 4
import os
import typer

from typer import echo
from tabulate import tabulate
import pandas as pd

from airt.cli import helper
from airt.logger import get_logger, set_level
from airt.constant import CLIENT_DB_USERNAME, CLIENT_DB_PASSWORD

# %% ../../notebooks/CLI_Pred.ipynb 6
app = typer.Typer(
    help="A set of commands for managing and downloading the predictions."
)

# %% ../../notebooks/CLI_Pred.ipynb 8
logger = get_logger(__name__)

# %% ../../notebooks/CLI_Pred.ipynb 12
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def ls(
    offset: int = typer.Option(
        0,
        "--offset",
        "-o",
        help="The number of predictions to offset at the beginning. If None, then the default value **0** will be used.",
    ),
    limit: int = typer.Option(
        100,
        "--limit",
        "-l",
        help="The maximum number of predictions to return from the server. If None, "
        "then the default value **100** will be used.",
    ),
    disabled: bool = typer.Option(
        False,
        "--disabled",
        help="If set to **True**, then only the deleted predictions will be returned. Else, the default value "
        "**False** will be used to return only the list of active predictions.",
    ),
    completed: bool = typer.Option(
        False,
        "--completed",
        help="If set to **True**, then only the predictions that are successfully downloaded "
        "to the server will be returned. Else, the default value **False** will be used to "
        "return all the predictions.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output only prediction uuids separated by space.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Return the list of predictions."""

    from airt.client import Prediction

    predx = Prediction.ls(
        offset=offset, limit=limit, disabled=disabled, completed=completed
    )

    df = Prediction.as_df(predx)

    df["created"] = helper.humanize_date(df["created"])

    return {"df": df, "quite_column_name": "prediction_uuid"}

# %% ../../notebooks/CLI_Pred.ipynb 20
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def details(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Return the details of a prediction."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)
    df = pred.details()

    df["created"] = helper.humanize_date(df["created"])

    return {"df": df}

# %% ../../notebooks/CLI_Pred.ipynb 24
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def rm(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output the deleted Prediction uuid only.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Delete a prediction from the server."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)
    df = pred.delete()

    df["created"] = helper.humanize_date(df["created"])

    return {"df": df, "quite_column_name": "prediction_uuid"}

# %% ../../notebooks/CLI_Pred.ipynb 28
@app.command("to-pandas")
@helper.requires_auth_token
def to_pandas(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Return the prediction results as a pandas DataFrame."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)

    df = pred.to_pandas()

    typer.echo(tabulate(df, headers="keys", tablefmt="plain", showindex=False))  # type: ignore

# %% ../../notebooks/CLI_Pred.ipynb 32
@app.command("to-s3")
@helper.requires_auth_token
def to_s3(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    uri: str = typer.Option(..., help="The target S3 bucket uri."),
    access_key: Optional[str] = typer.Option(
        None,
        help="Access key for the target S3 bucket. If **None** (default value), then the value from **AWS_ACCESS_KEY_ID** environment variable is used.",
    ),
    secret_key: Optional[str] = typer.Option(
        None,
        help="Secret key for the target S3 bucket. If **None** (default value), then the value from **AWS_SECRET_ACCESS_KEY** environment variable is used.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output status only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Push the prediction results to the target AWS S3 bucket."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)

    status = pred.to_s3(uri=uri, access_key=access_key, secret_key=secret_key)

    if quiet:
        status.wait()
        typer.echo(f"{pred.uuid}")
    else:
        typer.echo(
            f"Pushing the results for Prediction uuid: {pred.uuid} to the s3 bucket."
        )
        status.progress_bar()

# %% ../../notebooks/CLI_Pred.ipynb 36
@app.command("to-azure-blob-storage")
@helper.requires_auth_token
def to_azure_blob_storage(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    uri: str = typer.Option(..., help="Target Azure Blob Storage uri."),
    credential: str = typer.Option(
        ...,
        "--credential",
        "-c",
        help="Credential to access the Azure Blob Storage.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output status only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Push the prediction results to the target Azure Blob Storage."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)
    status = pred.to_azure_blob_storage(uri=uri, credential=credential)

    if quiet:
        status.wait()
        typer.echo(f"{pred.uuid}")
    else:
        typer.echo(
            f"Pushing the results for Prediction uuid: {pred.uuid} to the Azure Blob Storage."
        )
        status.progress_bar()

# %% ../../notebooks/CLI_Pred.ipynb 40
@app.command("to-local")
@helper.requires_auth_token
def to_local(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    path: str = typer.Option(..., help="Local directory path."),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output status only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Download the prediction results to a local directory."""

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)

    if quiet:
        pred.to_local(path=path, show_progress=False)
        typer.echo(f"{pred.uuid}")
    else:
        typer.echo(f"Downloading prediction results for uuid: {pred.uuid}.")
        pred.to_local(path=path)

# %% ../../notebooks/CLI_Pred.ipynb 44
@app.command("to-mysql")
@helper.requires_auth_token
def to_mysql(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    host: str = typer.Option(..., help="Database host name."),
    database: str = typer.Option(..., help="Database name."),
    table: str = typer.Option(..., help="Table name."),
    port: int = typer.Option(
        3306,
        help="Host port number. If not passed, then the default value **3306** will be used.",
    ),
    username: Optional[str] = typer.Option(
        None,
        "--username",
        "-u",
        help="Database username. If not passed, then the value set in the environment variable"
        f" **{CLIENT_DB_USERNAME}** will be used else the default value **root** will be used.",
    ),
    password: Optional[str] = typer.Option(
        None,
        "--password",
        "-p",
        help="Database password. If not passed, then the value set in the environment variable"
        f' **{CLIENT_DB_PASSWORD}** will be used else the default value "" will be used.',
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output status only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Push the prediction results to a mysql database.

    If the database requires authentication, pass the username/password as commandline arguments or store it in
    the **AIRT_CLIENT_DB_USERNAME** and **AIRT_CLIENT_DB_PASSWORD** environment variables.
    """

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)

    status = pred.to_mysql(
        host=host,
        database=database,
        table=table,
        port=port,
        username=username,
        password=password,
    )

    if quiet:
        status.wait()
        typer.echo(f"{pred.uuid}")
    else:
        typer.echo(
            f"Pushing the results for Prediction uuid: {pred.uuid} to the mysql database."
        )
        status.progress_bar()

# %% ../../notebooks/CLI_Pred.ipynb 47
@app.command("to-clickhouse")
@helper.requires_auth_token
def to_clickhouse(
    uuid: str = typer.Argument(
        ...,
        help="Prediction uuid.",
    ),
    host: str = typer.Option(..., help="Remote database host name."),
    database: str = typer.Option(..., help="Database name."),
    table: str = typer.Option(..., help="Table name."),
    protocol: str = typer.Option(..., help="Protocol to use (native/http)."),
    port: int = typer.Option(
        0,
        help="Host port number. If not passed, then the default value **0** will be used.",
    ),
    username: Optional[str] = typer.Option(
        None,
        "--username",
        "-u",
        help="Database username. If not passed, then the value set in the environment variable"
        " **CLICKHOUSE_USERNAME** will be used else the default value **root** will be used.",
    ),
    password: Optional[str] = typer.Option(
        None,
        "--password",
        "-p",
        help="Database password. If not passed, then the value set in the environment variable"
        ' **CLICKHOUSE_PASSWORD** will be used else the default value "" will be used.',
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output status only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Push the prediction results to a clickhouse database.

    If the database requires authentication, pass the username/password as commandline arguments or store it in
    the **CLICKHOUSE_USERNAME** and **CLICKHOUSE_PASSWORD** environment variables.
    """

    from airt.client import Prediction

    pred = Prediction(uuid=uuid)

    status = pred.to_clickhouse(
        host=host,
        database=database,
        table=table,
        port=port,
        protocol=protocol,
        username=username,
        password=password,
    )

    if quiet:
        status.wait()
        typer.echo(f"{pred.uuid}")
    else:
        typer.echo(
            f"Pushing the results for Prediction uuid: {pred.uuid} to the clickhouse database."
        )
        status.progress_bar()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/API_ProgressStatus.ipynb.

# %% auto 0
__all__ = ['ProgressStatus']

# %% ../../notebooks/API_ProgressStatus.ipynb 3
from typing import *

# %% ../../notebooks/API_ProgressStatus.ipynb 4
from time import sleep
from datetime import datetime, timedelta

from tqdm import tqdm

from fastcore.foundation import patch

from airt.components.client import Client
from airt.logger import get_logger, set_level

# %% ../../notebooks/API_ProgressStatus.ipynb 6
logger = get_logger(__name__)

# %% ../../notebooks/API_ProgressStatus.ipynb 9
class ProgressStatus:
    """A base class for querying status of a remote operation.

    Here's an example of using the ProgressStatus class to checking the upload status of the datablob

    Example:
        ```python
        # Importing necessary libraries
        from  airt.client import Client, DataBlob

        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a datablob
        # In this example, the datablob will be stored in an AWS S3 bucket. The region
        # is set to eu-west-3, feel free to change the cloud provider and the region
        # to suit your needs.
        db = DataBlob.from_s3(
            uri="{fill in uri}",
            cloud_provider="aws",
            region="eu-west-3"
        )

        # Check the ready status of the datablob
        print(db.is_ready())

        # Display the status in a progress bar
        db.progress_bar()

        # Check the ready status of the datablob
        # If the upload is successful, True will be returned
        print(db.is_ready())
        ```
    """

    def __init__(
        self, relative_url: str, sleep_for: Union[int, float] = 5, timeout: int = 0
    ):
        """Constructs a new ProgressStatus instance.

        Warning:
            Do not construct this object directly by calling the constructor, please use either progress_bar,
            is_ready, or wait methods of `DataBlob`, `DataSource`, `Model` or `Prediction` classes instead.

        Args:
            relative_url: Relative URI to query the status of the remote operation.
            sleep_for: The time interval in seconds between successive API calls.
            timeout: The maximum time allowed in seconds for the asynchronous call to complete. If not the
                progressbar will be terminated.

        Raises:
            TimeoutError: in case of connection timeout.
        """
        self.relative_url = relative_url
        self.sleep_for = sleep_for
        self.timeout = timeout

    def is_ready(self) -> bool:
        """Check if the method's progress is complete.

        Returns:
            **True** if the progress is completed, else **False**.

        Here's an example of checking the upload status of the datablob:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_s3(
                uri="{fill in uri}",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Check the ready status of the datablob
            print(db.is_ready())

            # Display the status in a progress bar
            db.progress_bar()

            # Check the ready status of the datablob
            # If the upload is successful, True will be returned
            print(db.is_ready())
            ```
        """
        response = Client._get_data(relative_url=self.relative_url)
        return response["completed_steps"] == response["total_steps"]

    def progress_bar(self):
        """Blocks the execution and displays a progress bar showing the remote action progress.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
            TimeoutError: in case of connection timeout.

        Here's an example of checking the upload status of the datablob:

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, DataBlob

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a datablob
            # In this example, the datablob will be stored in an AWS S3 bucket. The region
            # is set to eu-west-3, feel free to change the cloud provider and the region
            # to suit your needs.
            db = DataBlob.from_s3(
                uri="{fill in uri}",
                cloud_provider="aws",
                region="eu-west-3"
            )

            # Check the ready status of the datablob
            print(db.is_ready())

            # Display the status in a progress bar
            db.progress_bar()

            # Check the ready status of the datablob
            # If the upload is successful, True will be returned
            print(db.is_ready())
            ```
        """
        total_steps = Client._get_data(relative_url=self.relative_url)["total_steps"]
        with tqdm(total=total_steps) as pbar:
            started_at = datetime.now()
            while True:
                if (0 < self.timeout) and (datetime.now() - started_at) > timedelta(
                    seconds=self.timeout
                ):
                    raise TimeoutError()

                response = Client._get_data(relative_url=self.relative_url)
                completed_steps = response["completed_steps"]

                pbar.update(completed_steps)

                if completed_steps == total_steps:
                    break

                sleep(self.sleep_for)

    def wait(self):
        raise NotImplementedError()

# %% ../../notebooks/API_ProgressStatus.ipynb 16
@patch
def wait(self: ProgressStatus):
    """Blocks execution while waiting for the remote action to complete.

    Raises:
        ConnectionError: If the server address is invalid or not reachable.
        TimeoutError: in case of timeout.

    Here's an example of preventing further calls from being executed until the datablob upload is complete:

    Example:
        ```python
        # Importing necessary libraries
        from  airt.client import Client, DataBlob

        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a datablob
        # In this example, the datablob will be stored in an AWS S3 bucket. The region
        # is set to eu-west-3, feel free to change the cloud provider and the region
        # to suit your needs.
        db = DataBlob.from_s3(
            uri="{fill in uri}",
            cloud_provider="aws",
            region="eu-west-3"
        )

        # Further calls to the API will be blocked until the datablob upload is complete.
        db.wait()

        # Print the details of the newly created datablob
        # If the upload is successful, the ready flag should be set to True
        print(db.details())
        ```
    """
    started_at = datetime.now()
    while True:
        if (0 < self.timeout) and (datetime.now() - started_at) > timedelta(
            seconds=self.timeout
        ):
            raise TimeoutError()

        if self.is_ready():
            return

        sleep(self.sleep_for)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../notebooks/Logger.ipynb.

# %% auto 0
__all__ = ['should_supress_timestamps', 'supress_timestamps', 'get_logger', 'set_level']

# %% ../notebooks/Logger.ipynb 2
from typing import *
import logging

# %% ../notebooks/Logger.ipynb 3
import logging.config
from typer import echo

# %% ../notebooks/Logger.ipynb 4
logging_level = logging.WARNING

# %% ../notebooks/Logger.ipynb 6
# Logger Levels
# CRITICAL = 50
# ERROR = 40
# WARNING = 30
# INFO = 20
# DEBUG = 10
# NOTSET = 0

# Create a custom logger and set the configuration to echo the logging messages
# Reference: https://github.com/tiangolo/typer/issues/203


class CLILoggerHandler(logging.Handler):
    """A custom logger handler class that use Typer echo to log results."""

    def emit(self, record: logging.LogRecord) -> None:
        """A function to channel logger outout to Typer echo."""

        echo(self.format(record))


# Global handler for the loggers
handler = CLILoggerHandler()

# %% ../notebooks/Logger.ipynb 7
should_supress_timestamps: bool = False


def supress_timestamps(flag: bool = True):
    global should_supress_timestamps
    should_supress_timestamps = flag


def get_logger(name: str) -> logging.Logger:
    """An internal function to set the default logger configuration

    Returns:
        An instance of Logger with custom configuration that can be used across the CLI for logging
    """

    # Logger Configuration

    global should_supress_timestamps

    if should_supress_timestamps:
        fmt = "[%(levelname)s] %(name)s: %(message)s"
    else:
        fmt = "%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s: %(message)s"

    datefmt = "%y-%m-%d %H:%M:%S"

    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)

    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)

    return logger

# %% ../notebooks/Logger.ipynb 9
def set_level(level: int):

    # Getting all loggers that has either airt or __main__ in the name
    loggers = [
        logging.getLogger(name)
        for name in logging.root.manager.loggerDict
        if ("airt" in name) or ("__main__" in name)
    ]

    for logger in loggers:
        logger.setLevel(level)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/CLI_Helper.ipynb.

# %% auto 0
__all__ = ['logger', 'requires_auth_token', 'humanize_date', 'humanize_number', 'humanize_size', 'get_example_for_type',
           'get_example_output_format', 'customize_output_format', 'separate_integers_and_strings',
           'echo_formatted_output', 'display_formated_table', 'requires_totp', 'get_phone_registration_status']

# %% ../../notebooks/CLI_Helper.ipynb 3
import logging
from typing import *

# %% ../../notebooks/CLI_Helper.ipynb 4
import os
from contextlib import contextmanager
import functools
import ast
import datetime as dt

import pandas as pd
import typer
import humanize
from tabulate import tabulate

from airt.client import Client, User
from airt.logger import get_logger, set_level
from airt.constant import SERVER_URL, CLIENT_NAME, SERVICE_TOKEN, SERVICE_USERNAME

# %% ../../notebooks/CLI_Helper.ipynb 6
logger = get_logger(__name__)

# %% ../../notebooks/CLI_Helper.ipynb 12
@contextmanager
def authenicate_user():
    Client(
        auth_token=os.environ[SERVICE_TOKEN], server=os.environ.get(SERVER_URL, None)
    )

    yield

# %% ../../notebooks/CLI_Helper.ipynb 15
def requires_auth_token(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        try:

            if ("debug" in kwargs) and kwargs["debug"]:
                set_level(logging.DEBUG)
            else:
                set_level(logging.WARNING)

            with authenicate_user():
                # Do something before
                return func(*args, **kwargs)
                # Do something after

        except KeyError as e:
            typer.echo(
                message=f"KeyError: The environment variable {e} is not set.", err=True
            )

            if f"'{SERVICE_TOKEN}'" in str(e):
                typer.echo(
                    f"\nPlease run the command '{CLIENT_NAME} token' to get the application token and set it in the "
                    f"environment variable `{SERVICE_TOKEN}`."
                )
                typer.echo(f"\nTry '{CLIENT_NAME} token --help' for help.")

            raise typer.Exit(code=1)

        except Exception as e:
            typer.echo(message=f"Error: {e}", err=True)
            if ("Invalid OTP" in str(e)) or ("OTP is required" in str(e)):
                raise ValueError(e)
            raise typer.Exit(code=1)

    return wrapper_decorator

# %% ../../notebooks/CLI_Helper.ipynb 18
def humanize_date(s: pd.Series) -> pd.Series:
    return s.apply(
        lambda date: humanize.naturaltime(
            dt.datetime.now() - dt.datetime.strptime(date, "%Y-%m-%dT%H:%M:%S")  # type: ignore
        )
        if date
        else "None"
    )

# %% ../../notebooks/CLI_Helper.ipynb 20
def humanize_number(s: pd.Series) -> pd.Series:
    return s.apply(
        lambda num: humanize.intcomma(int(num)) if pd.notna(num) else "unknown"
    )

# %% ../../notebooks/CLI_Helper.ipynb 22
def humanize_size(s: pd.Series) -> pd.Series:
    return s.apply(
        lambda size: humanize.naturalsize(size) if pd.notna(size) else "unknown"
    )

# %% ../../notebooks/CLI_Helper.ipynb 24
def get_example_for_type(xs: pd.Series) -> str:
    """Get example output for the given series

    Args:
        xs: Input series

    Returns:
        The valid formatting example for the series
    """
    if pd.api.types.is_float_dtype(xs):
        return "€{:,.2f}"
    if pd.api.types.is_integer_dtype(xs):
        return "{:,d}"

    return "{}"

# %% ../../notebooks/CLI_Helper.ipynb 26
def get_example_output_format(df: pd.DataFrame) -> Dict[str, str]:
    """Get example output format for the dataframe

    Args:
        df: Input dataframe

    Returns:
        The example output format for the dataframe
    """
    return {c: get_example_for_type(df[c]) for c in df.columns}  # type: ignore

# %% ../../notebooks/CLI_Helper.ipynb 28
def customize_output_format(format_str: str, df: pd.DataFrame) -> pd.DataFrame:
    """Customize output format

    Args:
        format_str: A dict mapping of column names into their python format
        df: Input dataframe

    Returns:
        The formatted pandas DataFrame

    Raises:
        Error: If the formatting string is not a valid python expression
        Error: If invalid column name is passed
        Error: If invalid formatting string is passed
        Error: If wrong formatting is passed for a column
    """
    try:
        formatters = ast.literal_eval(format_str)
    except Exception as e:
        typer.echo(f"Not a valid python expression: {format_str}", err=True)
        typer.echo(
            f"An example of a valid formatting string: {get_example_output_format(df)}",
            err=True,
        )
        raise typer.Exit(code=1)

    if not isinstance(formatters, dict):
        typer.echo(f"The format string is not a dictionary: {formatters}", err=True)
        typer.echo(
            f"An example of a valid formatting string: {get_example_output_format(df)}",
            err=True,
        )
        raise typer.Exit(code=1)

    if not (set(formatters.keys()) <= set(df.columns)):
        typer.echo(
            f"The following columns are not valid: {set(formatters.keys()) - set(df.columns)}. Only the following columns are valid: {set(df.columns)}",
            err=True,
        )
        typer.echo(
            f"An example of a valid formatting string: {get_example_output_format(df)}",
            err=True,
        )
        raise typer.Exit(code=1)

    df_copy = df.copy()
    try:
        for k, v in formatters.items():
            df_copy[k] = df_copy[k].apply(lambda x: v.format(x))
    except Exception as e:
        typer.echo(f"Formatting is wrong for {k}: {v}", err=True)
        typer.echo(
            f"An example of a valid formatting string: {get_example_output_format(df)}",
            err=True,
        )
        raise typer.Exit(code=1)
    return df_copy[formatters.keys()]  # type: ignore

# %% ../../notebooks/CLI_Helper.ipynb 34
def separate_integers_and_strings(xs: List[str]) -> List[Union[int, str]]:
    """Seperate integers and strings from the list of strings

    Args:
        xs: List containing string inputs

    Returns:
        A list containing the integers and strings
    """
    return [int(v) if v.isdigit() else v for v in xs]

# %% ../../notebooks/CLI_Helper.ipynb 36
def echo_formatted_output(df: pd.DataFrame):
    """Echo the formatted output to the terminal

    Args:
        df: Input DataFrame
    """
    if len(df.columns) > 1:
        typer.echo(tabulate(df, headers="keys", tablefmt="plain", showindex=False))  # type: ignore
    else:
        single_col_results = df.iloc[:, 0].astype(str).to_list()
        typer.echo("\n".join(single_col_results))

# %% ../../notebooks/CLI_Helper.ipynb 40
def display_formated_table(func):
    """A decorator function to format the CLI table output"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):

        # Do something before
        result_dict = func(*args, **kwargs)
        # Do something after

        df = result_dict["df"]
        quite_column_name = (
            result_dict["quite_column_name"]
            if "quite_column_name" in result_dict
            else "uuid"
        )

        if kwargs["format"]:
            df = customize_output_format(kwargs["format"], df)
            echo_formatted_output(df)

        elif "quiet" in kwargs and kwargs["quiet"]:
            ids = df[quite_column_name].astype(str).to_list()
            typer.echo("\n".join(ids))

        else:
            typer.echo(
                tabulate(
                    result_dict["df"],
                    headers="keys",
                    tablefmt="plain",
                    showindex=False,
                    missingval="<none>",
                )
            )

    return wrapper

# %% ../../notebooks/CLI_Helper.ipynb 42
def requires_totp_or_otp(
    message_template_name: str, no_retries: int = 3, requires_auth_token: bool = True
):
    """A decorator function to prompt users to enter a valid totp or otp"""

    def wrapper_fn(f):
        @functools.wraps(f)
        def new_wrapper(*args, **kwargs):
            if requires_auth_token:
                try:
                    Client(
                        auth_token=os.environ[SERVICE_TOKEN],
                        server=os.environ.get(SERVER_URL, None),
                    )
                except KeyError as e:
                    typer.echo(
                        message=f"KeyError: The environment variable {e} is not set.",
                        err=True,
                    )

                    if f"'{SERVICE_TOKEN}'" in str(e):
                        typer.echo(
                            f"\nPlease run the command '{CLIENT_NAME} token' to get the application token and set it in the "
                            f"environment variable `{SERVICE_TOKEN}`."
                        )
                        typer.echo(f"\nTry '{CLIENT_NAME} token --help' for help.")

                    raise typer.Exit(code=1)

            # Non-Interactive mode
            if kwargs["otp"] is not None:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    raise typer.Exit(code=1)
            else:
                # Interactive mode
                if not requires_auth_token:
                    try:
                        return f(*args, **kwargs)
                    except ValueError as e:
                        pass
                typer.echo("\nPlease choose an option\n\n")
                while True:
                    typer.echo(
                        "[1] Use the dynamically generated six-digit verification code from the authenticator application\n"
                    )
                    typer.echo(
                        "[2] Request the OTP via SMS to the registered phone number\n"
                    )
                    typer.echo(
                        "If you cannot access the authenticator application and your registered phone number, please contact your administrator.\n"
                    )

                    user_option = typer.prompt("Enter your option")

                    if user_option in ["1", "2"]:
                        break
                    typer.echo("Please enter a valid option")

                if user_option == "1":
                    for i in range(no_retries):
                        _totp = typer.prompt(
                            "Please enter the OTP displayed in the authenticator app"
                        )
                        kwargs["otp"] = _totp
                        try:
                            return f(*args, **kwargs)
                        except ValueError as e:
                            pass
                    raise typer.Exit(code=1)
                else:
                    if requires_auth_token:
                        username = User.details()["username"]
                    else:
                        username = (
                            kwargs["username"]
                            if kwargs["username"] is not None
                            else os.environ.get(SERVICE_USERNAME)
                        )
                    sms_status = User.send_sms_otp(
                        username=username, message_template_name=message_template_name
                    )
                    typer.echo(f"\n{sms_status}\n")
                    for i in range(no_retries):
                        _sms_otp = typer.prompt(
                            f"Please enter the One-Time Password (OTP) you received on your registered phone number"
                        )
                        kwargs["otp"] = _sms_otp
                        try:
                            return f(*args, **kwargs)
                        except ValueError as e:
                            pass
                    raise typer.Exit(code=1)

        return new_wrapper

    return wrapper_fn

# %% ../../notebooks/CLI_Helper.ipynb 44
def requires_totp(no_retries: int = 3):
    """A decorator function to prompt users to enter a valid totp"""

    def wrapper_fn(f):
        @functools.wraps(f)
        def new_wrapper(*args, **kwargs):
            # Non-Interactive mode
            if kwargs["otp"] is not None:
                try:
                    return f(*args, **kwargs)
                except Exception as e:
                    raise typer.Exit(code=1)
            else:
                # Interactive mode
                _activation_otp = None
                for i in range(no_retries):
                    if _activation_otp is not None:
                        kwargs["otp"] = _activation_otp
                    try:
                        return f(*args, **kwargs)
                    except ValueError as e:
                        _activation_otp = typer.prompt(
                            "Please enter the OTP displayed in the authenticator app"
                        )
                raise typer.Exit(code=1)

        return new_wrapper

    return wrapper_fn

# %% ../../notebooks/CLI_Helper.ipynb 46
PHONE_REGISTRATION_STATUS = {
    "not_registered": (
        "\n\nPlease take a moment to register and verify your phone number. If you forget your password or cannot access your account, you can request "
        "the OTP to your registered phone number to regain access. \nTo register a new phone number, please set the token in the AIRT_SERVICE_TOKEN environment "
        f"variable and execute the below command with your phone number and follow the on-screen instructions:\n\n{CLIENT_NAME} user register-phone-number --phone-number"
    ),
    "not_validated": (
        "\n\nYour phone number is added to your account but not yet verified. Please take a moment to register and verify your phone number. If you forget "
        "your password or cannot access your account, you can request the OTP to your registered phone number to regain access. \nTo register a new phone number, please set "
        "the token in the AIRT_SERVICE_TOKEN environment variable and execute the below command with your phone number and follow the on-screen instructions:"
        f"\n\n{CLIENT_NAME} user register-phone-number"
        "\n\nIn case you want to register a new number, please execute the below command with your new phone number:"
        f"\n\n{CLIENT_NAME} user register-phone-number --phone-number"
    ),
}

# %% ../../notebooks/CLI_Helper.ipynb 47
def get_phone_registration_status(xs: Dict[str, Union[str, bool]]) -> Optional[str]:
    """Get the phone number registration status

    Args:
        xs: A dict containing the user details

    Returns:
        None, if the user phone number is registred and validated. Else, a message containing the current state of the phone number registration process.
    """
    if not xs["phone_number"]:
        return PHONE_REGISTRATION_STATUS["not_registered"]
    if not xs["is_phone_number_verified"]:
        return PHONE_REGISTRATION_STATUS["not_validated"]
    return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/API_Keys.ipynb.

# %% auto 0
__all__ = ['APIKey']

# %% ../../notebooks/API_Keys.ipynb 4
from typing import *

# %% ../../notebooks/API_Keys.ipynb 5
import os

from datetime import datetime, timedelta, date

import pandas as pd
from fastcore.foundation import patch

from airt.components.client import Client
from airt.components.user import User
from airt.logger import get_logger, set_level
from airt.helper import (
    get_data,
    post_data,
    delete_data,
    generate_df,
    get_attributes_from_instances,
    check_and_append_otp_query_param,
)

# %% ../../notebooks/API_Keys.ipynb 7
logger = get_logger(__name__)

# %% ../../notebooks/API_Keys.ipynb 11
class APIKey:
    """A class for managing the APIKeys in the server.

    Both the APIKey and the token can be used for accessing the airt services. However, there is a slight difference in generating and managing the two.

    For generating the APIKey, you first need to get the developer token. Please refer to `Client.get_token` method documentation to generate one.

    After logging in with your developer token, you can create any number of new APIKeys and can set an expiration date individually. You can also access
    other methods available in the APIKey class to list, revoke the APIKey at any time.

    Here's an example of how to use the APIKey class to create a new key and use it to access the airt service.

    Example:
        ```python
        # Importing necessary libraries
        from  airt.client import Client, APIKey, User

        # Authenticate
        Client.get_token(username="{fill in username}", password="{fill in password}")

        # Create a new key with the given name
        key_name = "{fill in key_name}"
        new_key = APIKey.create(name=key_name)

        # Display the details of the newly created key
        print(APIKey.details(apikey=key_name))

        # Call the set_token method to set the newly generated key
        Client.set_token(token=new_key["access_token"])

        # Print the logged-in user details
        # If set_token fails, the line below will throw an error.
        print(User.details())
        ```
    """

    API_KEY_COLS = ["uuid", "name", "created", "expiry", "disabled"]

    def __init__(
        self,
        uuid: str,
        name: Optional[str] = None,
        expiry: Optional[str] = None,
        disabled: Optional[bool] = None,
        created: Optional[str] = None,
    ):
        """Constructs a new APIKey instance.

        Args:
            uuid: APIKey uuid.
            name: APIKey name.
            expiry: APIKey expiry date.
            disabled: Flag to indicate the status of the APIKey.
            created: APIKey creation date.
        """
        self.uuid = uuid
        self.name = name
        self.expiry = expiry
        self.disabled = disabled
        self.created = created

    @staticmethod
    def create(
        name: str,
        expiry: Optional[Union[int, timedelta, datetime]] = None,
        otp: Optional[str] = None,
    ) -> Dict[str, str]:
        """Create a new APIKey

        In order to access the airt service with the newly generated APIKey, please call the `Client.set_token` method
        or set the APIKey value in the **AIRT_SERVICE_TOKEN** environment variable.

        !!! note

            - The APIKey's name must be unique. If not, an exception will be raised while creating a new key with the name of an existing key.
            However, you can create a new key with the name of a revoked key.

            - The expiry for an APIKey is optional, if not passed then the default value **None** will be used to create an APIKey with no expiry date!

        Args:
            name: The name of the APIKey.
            expiry: The validity for the APIKey. This can be an integer representing the number of days till expiry, can be
                an instance of timedelta (timedelta(days=x)) or can be an instance of datetime. If not passed, then the default value
                **None** will be used to create a APIKey that will never expire!
            otp: Dynamically generated six-digit verification code from the authenticator app. Please pass this
                parameter only if the MFA is enabled for your account.

        Returns:
            The APIKey and its type as a dictionary.

        Raises:
            ValueError: If the user is not authenticated.
            ValueError: If the user tries to create a new APIKey with an existing key name.
            ValueError: If the OTP is invalid.
            ConnectionError: If the server address is invalid or not reachable.

        In the following example, a new APIKey is created with a 10-day expiration date and used to access the airt service.

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a key with the given name and set the expiry to 10 days from now.
            # If the expiry parameter is not specified, a key with no expiry date is created.
            key_name = "{fill in key_name}"
            new_key_details = APIKey.create(name=key_name, expiry=10)

            # Display the details of the newly created key
            print(APIKey.details(apikey=key_name))

            # If a new key with the same name is created, an exception will be raised.
            # However, you can create a new key with the name of a revoked key.
            try:
                APIKey.create(name=key_name, expiry=10)
                print("Should not print this, the above line should raise an exception")
                raise RuntimeException()

            except ValueError as e:
                print("Expected to fail, everything is fine")

            # Finally, either call the below method to set the newly generated key
            # or store it in the AIRT_SERVICE_TOKEN environment variable.
            Client.set_token(token=new_key_details["access_token"])

            # If set_token fails, the line below will throw an error.
            print(APIKey.details(apikey=key_name))
            ```
        """
        if expiry is None:
            expiry_date = expiry
        else:
            if isinstance(expiry, int):
                delta = datetime.now() + timedelta(days=expiry)
            elif isinstance(expiry, timedelta):
                delta = datetime.now() + expiry
            else:
                delta = expiry

            expiry_date = delta.strftime("%Y-%m-%dT%H:%M")

        return Client._post_data(
            relative_url="/apikey",
            json=dict(name=name, expiry=expiry_date, otp=otp),
        )

    @staticmethod
    def as_df(ax: List["APIKey"]) -> pd.DataFrame:
        """Return the details of APIKey instances in a pandas dataframe.

        Args:
            ax: List of APIKey instances.

        Returns:
            Details of all the APIKeys in a dataframe.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.

        An example of displaying the APIKeys generated by the currently logged-in user in a dataframe

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a key without an expiry date in the given name
            key_name = "{fill in key_name}"
            APIKey.create(name=key_name)

            # Display all the APIKey instance details in a pandas dataframe
            df = APIKey.as_df(APIKey.ls())
            print(df)
            ```
        """
        lists = get_attributes_from_instances(ax, APIKey.API_KEY_COLS)  # type: ignore
        return generate_df(lists, APIKey.API_KEY_COLS)

    @staticmethod
    def ls(
        user: Optional[str] = None,
        offset: int = 0,
        limit: int = 100,
        include_disabled: bool = False,
    ) -> List["APIKey"]:
        """Return the list of APIKeys instances.

        Please do not pass the **user** parameter unless you are a super user. Only a super user can view
        the APIKeys created by other users.

        Args:
            user: user_uuid/username associated with the APIKey. Please call `User.details` method of the User class to get your user_uuid.
                If not passed, then the currently logged-in user_uuid will be used.
            offset: The number of APIKeys to offset at the beginning. If None, then the default value 0 will be used.
            limit: The maximum number of APIKeys to return from the server. If None, then the default value 100 will be used.
            include_disabled: If set to **True**, then the disabled APIKeys will also be included in the result.

        Returns:
            A list of APIKey instances.

        Raises:
            ConnectionError: If the server address is invalid or not reachable.
            ValueError: If the user_uuid is invalid.

        An example of displaying the APIKeys generated by the currently logged-in user

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a key without an expiry date in the given name
            key_name = "{fill in key_name}"
            APIKey.create(name=key_name)

            # Get the list of all APIKey instances created by the currently logged-in user.
            # If you are a super user, you can view the APIkeys created by other users by
            # passing their uuid/username in the user parameter.
            ax = APIKey.ls()
            print(ax)

            # Display the details of the instances in a pandas dataframe
            df = APIKey.as_df(ax)
            print(df)
            ```
        """
        user_uuid = User.details(user=user)["uuid"]

        apikeys = Client._get_data(
            relative_url=f"/{user_uuid}/apikey?include_disabled={include_disabled}&offset={offset}&limit={limit}"
        )

        ax = [
            APIKey(
                uuid=apikey["uuid"],
                name=apikey["name"],
                expiry=apikey["expiry"],
                disabled=apikey["disabled"],
                created=apikey["created"],
            )
            for apikey in apikeys
        ]

        return ax

    @staticmethod
    def details(apikey: str) -> pd.DataFrame:
        """Return details of an APIKey.

        Args:
            apikey: APIKey uuid/name.

        Returns:
            A pandas Dataframe encapsulating the details of the APIKey.

        Raises:
            ValueError: If the APIKey uuid is invalid.
            ConnectionError: If the server address is invalid or not reachable.

        An example to get details of an APIKey

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a key without an expiry date in the given name
            key_name = "{fill in key_name}"
            APIKey.create(name=key_name)

            # Display the details of the newly created key
            print(APIKey.details(apikey=key_name))

            # To display the details of all keys created by the user, use the method below.
            df = APIKey.as_df(APIKey.ls())
            print(df)
            ```
        """
        details = Client._get_data(relative_url=f"/apikey/{apikey}")

        return pd.DataFrame(details, index=[0])[APIKey.API_KEY_COLS]

    @staticmethod
    def _get_key_names(keys: Union[List["APIKey"], List[str], str]) -> List[str]:
        """Get keys names

        Args:
            keys: Can be a string, list of string or a list of APIKey instances

        Returns:
            The APIKey names as a list
        """
        if isinstance(keys, str):
            return [keys]

        if all(isinstance(k, str) for k in keys):
            return keys  # type: ignore

        return [k.name for k in keys]  # type: ignore

    @staticmethod
    def revoke(
        keys: Union[str, List[str], List["APIKey"]],
        user: Optional[str] = None,
        otp: Optional[str] = None,
    ) -> pd.DataFrame:
        """Revoke one or more APIKeys

        Please do not pass the **user** parameter unless you are a super user. Only a super user can revoke the
        APIKeys created by other users.

        Args:
            keys: APIKey uuid/name to revoke. To revoke multiple keys, either pass a list of APIKey uuid/names or a list of APIKey instances.
            user: user_uuid/username associated with the APIKey. Please call `User.details` method of the User class to get your user_uuid/username.
                If not passed, then the currently logged-in user will be used.
            otp: Dynamically generated six-digit verification code from the authenticator app. Please pass this
                parameter only if the MFA is enabled for your account.

        Returns:
             A pandas Dataframe encapsulating the details of the deleted APIKey(s).

        Raises:
            ValueError: If the APIKey uuid is invalid.
            ValueError: If the user_uuid is invalid.
            ValueError: If the OTP is invalid.
            ConnectionError: If the server address is invalid or not reachable.

        An example to revoke a single APIKey by name

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate
            Client.get_token(username="{fill in username}", password="{fill in password}")

            # Create a key without an expiry date in the given name
            key_name = "{fill in key_name}"
            APIKey.create(name=key_name)

            # Check that the newly created key exists
            print([key.name for key in APIKey.ls()])

            # Revoke the newly created key
            # To delete multiple keys, pass a list of key names or key instances
            APIKey.revoke(keys=key_name)

            # Check that the newly created key does not exists
            print([key.name for key in APIKey.ls()])
            ```

        Here's an example of a super user revoking all APIkeys generated by a specific user.

        Example:
            ```python
            # Importing necessary libraries
            from  airt.client import Client, APIKey

            # Authenticate with super user privileges
            Client.get_token(
                username="{fill in super_user_username}",
                password="{fill in super_user_password}"
            )

            # List the APIKeys generated by a specific user
            user = "{fill in other_username}"
            ax = APIKey.ls(user=user)
            print([key.name for key in ax])

            # Revoke the APIKeys
            APIKey.revoke(keys=ax, user=user)

            # Check that all APIkeys have been revoked
            print([key.name for key in APIKey.ls(user=user)])
            ```
        """
        user_uuid = User.details(user=user)["uuid"]
        _keys = APIKey._get_key_names(keys)

        response_list = []

        for key_uuid in _keys:
            url = f"/{user_uuid}/apikey/{key_uuid}"
            response = Client._delete_data(
                relative_url=check_and_append_otp_query_param(url, otp)
            )
            response_list.append(response)

        return generate_df(response_list, APIKey.API_KEY_COLS)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../notebooks/CLI_DataSource.ipynb.

# %% auto 0
__all__ = ['logger']

# %% ../../notebooks/CLI_DataSource.ipynb 3
from typing import *

# %% ../../notebooks/CLI_DataSource.ipynb 4
import os

import typer
from typer import echo
from tabulate import tabulate
import datetime as dt
import pandas as pd

from airt.client import Client
from airt.cli import helper
from airt.logger import get_logger, set_level

# %% ../../notebooks/CLI_DataSource.ipynb 6
app = typer.Typer(
    help="A set of commands for managing datasources and training ML models on them."
)

# %% ../../notebooks/CLI_DataSource.ipynb 8
logger = get_logger(__name__)

# %% ../../notebooks/CLI_DataSource.ipynb 14
@app.command()
@helper.requires_auth_token
def dtypes(
    uuid: str = typer.Argument(
        ...,
        help="Datasource uuid.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> None:
    """Return the dtypes of the datasource."""

    from airt.client import DataSource

    ds = DataSource(uuid=uuid)
    dtypes = ds.dtypes.T.rename(columns={0: "dtype"})
    typer.echo(dtypes)

# %% ../../notebooks/CLI_DataSource.ipynb 19
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def head(
    uuid: str = typer.Argument(
        ...,
        help="Datasource uuid.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Return the first few rows of the datasource."""

    from airt.client import DataSource

    ds = DataSource(uuid=uuid)
    df = ds.head()

    return {"df": df}

# %% ../../notebooks/CLI_DataSource.ipynb 22
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def ls(
    offset: int = typer.Option(
        0,
        "--offset",
        "-o",
        help="The number of datasources to offset at the beginning. If **None**, then the default value **0** will be used.",
    ),
    limit: int = typer.Option(
        100,
        "--limit",
        "-l",
        help="The maximum number of datasources to return from the server. If **None**, then the default value **100** will be used.",
    ),
    disabled: bool = typer.Option(
        False,
        "--disabled",
        help="If set to **True**, then only the deleted datasources will be returned."
        "Else, the default value **False** will be used to return only the list"
        "of active datasources.",
    ),
    completed: bool = typer.Option(
        False,
        "--completed",
        help="If set to **True**, then only the datasources that are successfully downloaded"
        "to the server will be returned. Else, the default value **False** will be used to"
        "return all the datasources.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output only datasource uuids separated by space",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Return the list of datasources."""

    from airt.client import DataSource

    dsx = DataSource.ls(
        offset=offset, limit=limit, disabled=disabled, completed=completed
    )
    df = DataSource.as_df(dsx)

    df["pulled_on"] = helper.humanize_date(df["pulled_on"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    return {"df": df, "quite_column_name": "datasource_uuid"}

# %% ../../notebooks/CLI_DataSource.ipynb 27
@app.command()
@helper.requires_auth_token
def train(
    uuid: str = typer.Option(
        ...,
        "--datasource_uuid",
        "-uuid",
        help="Datasource uuid.",
    ),
    client_column: str = typer.Option(
        ...,
        "--client_column",
        help="The column name that uniquely identifies the users/clients.",
    ),
    timestamp_column: Optional[str] = typer.Option(
        None,
        "--timestamp_column",
        help="The timestamp column indicating the time of an event. If not passed, then the default value **None** will be used. ",
    ),
    target_column: str = typer.Option(
        ...,
        "--target_column",
        help="Target column name that indicates the type of the event.",
    ),
    target: str = typer.Option(
        ...,
        "--target",
        help="Target event name to train and make predictions. You can pass the target event as a string or as a "
        "regular expression for predicting more than one event. For example, passing ***checkout** will "
        "train a model to predict any checkout event.",
    ),
    predict_after: str = typer.Option(
        ...,
        "--predict_after",
        help="Time delta in hours of the expected target event.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output model uuid only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
):
    """Train a model against the datasource."""

    from airt.client import DataSource
    from datetime import timedelta

    ds = DataSource(uuid=uuid)

    model = ds.train(
        client_column=client_column,
        target_column=target_column,
        target=target,
        predict_after=timedelta(hours=int(predict_after)),
        timestamp_column=timestamp_column,
    )

    if quiet:
        model.wait()
        typer.echo(f"{model.uuid}")
    else:
        typer.echo(f"Training started for model uuid: {model.uuid}")
        model.progress_bar()

# %% ../../notebooks/CLI_DataSource.ipynb 32
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def details(
    uuid: str = typer.Argument(
        ...,
        help="Datasource uuid.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Return details of a datasource."""

    from airt.client import DataSource

    ds = DataSource(uuid=uuid)
    df = ds.details()

    df["pulled_on"] = helper.humanize_date(df["pulled_on"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    return {"df": df}

# %% ../../notebooks/CLI_DataSource.ipynb 36
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def rm(
    uuid: str = typer.Argument(
        ...,
        help="Datasource uuid.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        help="Output the deleted datasource uuid only.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Delete a datasource from the server."""

    from airt.client import DataSource

    ds = DataSource(uuid=uuid)
    df = ds.delete()

    df["pulled_on"] = helper.humanize_date(df["pulled_on"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    return {"df": df, "quite_column_name": "datasource_uuid"}

# %% ../../notebooks/CLI_DataSource.ipynb 40
@app.command()
@helper.display_formated_table
@helper.requires_auth_token
def tag(
    uuid: str = typer.Option(
        ...,
        "--datasource_uuid",
        "-uuid",
        help="Datasource uuid.",
    ),
    name: str = typer.Option(
        ...,
        "--name",
        "-n",
        help="A string to tag the datasource.",
    ),
    format: Optional[str] = typer.Option(
        None,
        "--format",
        "-f",
        help="Format output and show only the given column(s) values.",
    ),
    debug: bool = typer.Option(
        False,
        "--debug",
        "-d",
        help="Set logger level to DEBUG and output everything.",
    ),
) -> Dict["str", Union[pd.DataFrame, str]]:
    """Tag an existing datasource in server."""

    from airt.client import DataSource

    ds = DataSource(uuid=uuid)
    df = ds.tag(name=name)

    df["pulled_on"] = helper.humanize_date(df["pulled_on"])
    df["no_of_rows"] = helper.humanize_number(df["no_of_rows"])
    df["folder_size"] = helper.humanize_size(df["folder_size"])

    return {"df": df}
